import torch

from .galaxy_model_object import Galaxy_Model
from .warp_model import Warp_Galaxy
from ..utils.decorators import default_internal

__all__ = ["SuperEllipse_Galaxy", "SuperEllipse_Warp"]


class SuperEllipse_Galaxy(Galaxy_Model):
    """Expanded galaxy model which includes a superellipse transformation
    in its radius metric. This allows for the expression of "boxy" and
    "disky" isophotes instead of pure ellipses. This is a common
    extension of the standard elliptical representation, especially
    for early-type galaxies. The functional form for this is:

    R = (|X|^C + |Y|^C)^(1/C)

    where R is the new distance metric, X Y are the coordinates, and C
    is the coefficient for the superellipse. C can take on any value
    greater than zero where C = 2 is the standard distance metric, 0 <
    C < 2 creates disky or pointed perturbations to an ellipse, and C
    > 2 transforms an ellipse to be more boxy.

    Parameters:
        C0: superellipse distance metric parameter where C0 = C-2 so that a value of zero is now a standard ellipse.

    """

    model_type = f"superellipse {Galaxy_Model.model_type}"
    parameter_specs = {
        "C0": {"units": "C-2", "value": 0.0, "uncertainty": 1e-2, "limits": (-2, None)},
    }
    _parameter_order = Galaxy_Model._parameter_order + ("C0",)
    usable = False

    @default_internal
    def radius_metric(self, X, Y, image=None, parameters=None):
        return torch.pow(
            torch.pow(torch.abs(X), parameters["C0"].value + 2.0)
            + torch.pow(torch.abs(Y), parameters["C0"].value + 2.0),
            1.0 / (parameters["C0"].value + 2.0),
        )


class SuperEllipse_Warp(Warp_Galaxy):
    """Expanded warp model which includes a superellipse transformation
    in its radius metric. This allows for the expression of "boxy" and
    "disky" isophotes instead of pure ellipses. This is a common
    extension of the standard elliptical representation, especially
    for early-type galaxies. The functional form for this is:

    R = (|X|^C + |Y|^C)^(1/C)

    where R is the new distance metric, X Y are the coordinates, and C
    is the coefficient for the superellipse. C can take on any value
    greater than zero where C = 2 is the standard distance metric, 0 <
    C < 2 creates disky or pointed perturbations to an ellipse, and C
    > 2 transforms an ellipse to be more boxy.

    Parameters:
        C0: superellipse distance metric parameter where C0 = C-2 so that a value of zero is now a standard ellipse.


    """

    model_type = f"superellipse {Warp_Galaxy.model_type}"
    parameter_specs = {
        "C0": {"units": "C-2", "value": 0.0, "uncertainty": 1e-2, "limits": (-2, None)},
    }
    _parameter_order = Warp_Galaxy._parameter_order + ("C0",)
    usable = False

    @default_internal
    def radius_metric(self, X, Y, image=None, parameters=None):
        return torch.pow(
            torch.pow(torch.abs(X), parameters["C0"].value + 2.0)
            + torch.pow(torch.abs(Y), parameters["C0"].value + 2.0),
            1.0 / (parameters["C0"].value + 2.0),
        )  # epsilon added for numerical stability of gradient



import numpy as np
from scipy.stats import iqr
import torch

from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..param import Param_Unlock, Param_SoftLimits
from .sky_model_object import Sky_Model
from ._shared_methods import select_target

__all__ = ["Flat_Sky"]


class Flat_Sky(Sky_Model):
    """Model for the sky background in which all values across the image
    are the same.

    Parameters:
        sky: brightness for the sky, represented as the log of the brightness over pixel scale squared, this is proportional to a surface brightness

    """

    model_type = f"flat {Sky_Model.model_type}"
    parameter_specs = {
        "F": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Sky_Model._parameter_order + ("F",)
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        with Param_Unlock(parameters["F"]), Param_SoftLimits(parameters["F"]):
            if parameters["F"].value is None:
                parameters["F"].value = torch.log10(
                    torch.abs(torch.median(target[self.window].data)) / target.pixel_area
                )
            if parameters["F"].uncertainty is None:
                parameters["F"].uncertainty = (
                    (
                        iqr(
                            target[self.window].data.detach().cpu().numpy(),
                            rng=(31.731 / 2, 100 - 31.731 / 2),
                        )
                        / (2.0 * target.pixel_area.item())
                    )
                    / np.sqrt(np.prod(self.window.shape.detach().cpu().numpy()))
                ) / (10 ** parameters["F"].value.item() * np.log(10))

    def evaluate_model(self, X=None, Y=None, image=None, parameters=None, **kwargs):
        ref = image.data if X is None else X
        return torch.ones_like(ref) * (image.pixel_area * 10 ** parameters["F"].value)



import numpy as np
import matplotlib.pyplot as plt
from astropy.visualization.mpl_normalize import ImageNormalize
from astropy.visualization import LogStretch, HistEqStretch


def LSBImage(dat, noise):
    plt.figure(figsize=(6, 6))
    plt.imshow(
        dat,
        origin="lower",
        cmap="Greys",
        norm=ImageNormalize(
            stretch=HistEqStretch(dat[dat <= 3 * noise]),
            clip=False,
            vmax=3 * noise,
            vmin=np.min(dat),
        ),
    )
    my_cmap = copy(cm.Greys_r)
    my_cmap.set_under("k", alpha=0)

    plt.imshow(
        np.ma.masked_where(dat < 3 * noise, dat),
        origin="lower",
        cmap=my_cmap,
        norm=ImageNormalize(stretch=LogStretch(), clip=False),
        clim=[3 * noise, None],
        interpolation="none",
    )
    plt.xticks([])
    plt.yticks([])
    plt.subplots_adjust(left=0.03, right=0.97, top=0.97, bottom=0.05)
    plt.xlim([0, dat.shape[1]])
    plt.ylim([0, dat.shape[0]])


def _display_time(seconds):
    intervals = (
        ("hours", 3600),  # 60 * 60
        ("arcminutes", 60),
        ("arcseconds", 1),
    )
    result = []

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip("s")
            result.append("{} {}".format(value, name))
    return ", ".join(result)


def AddScale(ax, img_width, loc="lower right"):
    """
    ax: figure axis object
    img_width: image width in arcseconds
    loc: location to put the scale bar
    """
    scale_width = int(img_width / 6)

    if scale_width > 60 and scale_width % 60 <= 15:
        scale_width -= scale_width % 60
    if scale_width > 45 and scale_width % 60 >= 45:
        scale_width += 60 - (scale_width % 60)
    if 15 < scale_width % 60 < 45:
        scale_width += 30 - (scale_width % 60)

    label = _display_time(scale_width)

    xloc = 0.05 if "left" in loc else 0.95
    yloc = 0.95 if "upper" in loc else 0.05

    ax.text(
        xloc - 0.5 * scale_width / img_width,
        yloc + 0.005,
        label,
        horizontalalignment="center",
        verticalalignment="bottom",
        transform=ax.transAxes,
        fontsize="x-small" if len(label) < 20 else "xx-small",
        weight="bold",
        color=autocolours["red1"],
    )
    ax.plot(
        [xloc - scale_width / img_width, xloc],
        [yloc, yloc],
        transform=ax.transAxes,
        color=autocolours["red1"],
    )


def AddLogo(fig, loc=[0.8, 0.01, 0.844 / 5, 0.185 / 5], white=False):
    im = plt.imread(
        get_sample_data(
            os.path.join(
                os.environ["AUTOPROF"],
                "_static/",
                ("AP_logo_white.png" if white else "AP_logo.png"),
            )
        )
    )
    newax = fig.add_axes(loc, zorder=1000)
    if white:
        newax.imshow(np.zeros(im.shape) + np.array([0, 0, 0, 1]))
    else:
        newax.imshow(np.ones(im.shape))
    newax.imshow(im)
    newax.axis("off")



from . import (
    optimization,
    angle_operations,
    decorators,
    interpolate,
    operations,
    parametric_profiles,
    isophote,
    initialize,
    conversions,
)

__all__ = [
    "optimization",
    "angle_operations",
    "decorators",
    "interpolate",
    "operations",
    "parametric_profiles",
    "isophote",
    "initialize",
    "conversions",
]



__all__ = ["galfit_config"]

galfit_object_type_map = {
    "sersic": "sersic galaxy model",
    "sky": "flat sky model",
}

galfit_parameter_map = {
    "sersic galaxy model": {
        "1": ["centerpix", 2],
        "3": ["totalmag", 1],
        "4": ["Repix", 1],
        "5": ["n", 1],
        "9": ["q", 1],
        "10": ["PAdeg", 1],
    }
}


def space_split(l):
    items = list(ls.strip() for ls in l.split(" "))
    index = 0
    while index < len(items):
        if items[index] == "":
            items.pop(index)
        else:
            index += 1
    return items


def galfit_config(config_file):
    if True:
        raise NotImplementedError("galfit configuration file interface under construction")
    with open(config_file, "r") as f:
        config_lines = f.readlines()
    # Header info
    headerinfo = {}
    for line in config_lines:
        # remove comment from line and strip whitespace
        comment = line.find("#")
        if comment >= 0:
            line = line[:comment].strip()
        if line == "":
            continue
        if line.startswith("A)"):
            headerinfo["target_file"] = line[2:].strip()
        if line.startswith("B)"):
            headerinfo["saveto_model"] = line[2:].strip()
        if line.startswith("C)"):
            headerinfo["varaince_file"] = line[2:].strip()
        if line.startswith("D)"):
            headerinfo["psf_file"] = line[2:].strip()
        if line.startswith("E)"):
            headerinfo["psf_upample"] = line[2:].strip()
        if line.startswith("F)"):
            headerinfo["mask_file"] = line[2:].strip()
        if line.startswith("G)"):
            headerinfo["constraints_file"] = line[2:].strip()
        if line.startswith("H)"):
            headerinfo["fit_window"] = line[2:].strip()
        if line.startswith("I)"):
            headerinfo["convolution_window"] = line[2:].strip()
        if line.startswith("J)"):
            headerinfo["target_zeropoint"] = line[2:].strip()
        if line.startswith("K)"):
            headerinfo["target_pixelscale"] = line[2:].strip()

    # Object info
    objects = []
    in_object = False
    for line in config_lines:
        # remove comment from line and strip whitespace
        comment = line.find("#")
        if comment >= 0:
            linem = line[:comment].strip()
        if linem == "":
            continue

        # New model added to the fit
        if linem.startswith("0)"):
            objects.append({"model_type": galfit_object_type_map[linem[2:].strip()]})
            in_object = True
        # Model finished adding
        if linem.startswith("Z)"):
            in_object = False

        # Collect the parameters
        if in_object:
            param = linem[: linem.find(")")]
            objects[-1][galfit_parameter_map[objects[-1]["model_type"]][param][0]] = space_split(
                linem[linem.find(")") + 1 :]
            )
            if len(objects[-1][galfit_parameter_map[objects[-1]["model_type"]][param][0]]) != (
                2 * galfit_parameter_map[objects[-1]["model_type"]][param][1]
            ):
                raise ValueError(f"Incorrectly formatted line in GALFIT config file:\n{line}")

    # Format parameters
    for i in range(len(objects)):
        astrophot_object = {
            "model_type": objects[i]["model_type"],
        }

        # common params
        if "centerpix" in objects[i]:
            astrophot_object["center"] = {
                "value": [
                    float(objects[i]["centerpix"][0]) * headerinfo["target_pixelscale"],
                    float(objects[i]["centerpix"][1]) * headerinfo["target_pixelscale"],
                ],
                "locked": bool(objects[i]["centerpix"][2]),
            }
        if "Repix" in objects[i]:
            astrophot_object["Re"] = {
                "value": float(objects[i]["Repix"][0]) * headerinfo["target_pixelscale"],
                "locked": bool(objects[i]["Repix"][1]),
            }
        if "q" in objects[i]:
            astrophot_object["q"] = {
                "value": float(objects[i]["q"][0]),
                "locked": bool(objects[i]["q"][1]),
            }
        if "PAdeg" in objects[i]:
            astrophot_object["PA"] = {
                "value": float(objects[i]["PAdeg"][0]) * np.pi / 180,
                "locked": bool(objects[i]["PAdeg"][1]),
            }



from functools import lru_cache

import torch
from scipy.special import binom

from ..utils.decorators import ignore_numpy_warnings, default_internal
from ._shared_methods import select_target
from .psf_model_object import PSF_Model
from ..param import Param_Unlock, Param_SoftLimits
from ..errors import SpecificationConflict

__all__ = ("Zernike_PSF",)


class Zernike_PSF(PSF_Model):

    model_type = f"zernike {PSF_Model.model_type}"
    parameter_specs = {
        "Anm": {"units": "flux/arcsec^2"},
    }
    _parameter_order = PSF_Model._parameter_order + ("Anm",)
    usable = True
    model_integrated = False

    def __init__(self, *, name=None, order_n=2, r_scale=None, **kwargs):
        super().__init__(name=name, **kwargs)

        self.order_n = int(order_n)
        self.r_scale = r_scale
        self.nm_list = self.iter_nm(self.order_n)

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        # List the coefficients to use
        self.nm_list = self.iter_nm(self.order_n)
        # Set the scale radius for the Zernike area
        if self.r_scale is None:
            self.r_scale = torch.max(self.window.shape) / 2

        # Check if user has already set the coefficients
        if parameters["Anm"].value is not None:
            if len(self.nm_list) != len(parameters["Anm"].value):
                raise SpecificationConflict(
                    f"nm_list length ({len(self.nm_list)}) must match coefficients ({len(parameters['Anm'].value)})"
                )
            return

        # Set the default coefficients to zeros
        with Param_Unlock(parameters["Anm"]), Param_SoftLimits(parameters["Anm"]):
            parameters["Anm"].value = torch.zeros(len(self.nm_list))
            if parameters["Anm"].uncertainty is None:
                parameters["Anm"].uncertainty = self.default_uncertainty * torch.ones_like(
                    parameters["Anm"].value
                )
            # Set the zero order zernike polynomial to the average in the image
            if self.nm_list[0] == (0, 0):
                parameters["Anm"].value[0] = (
                    torch.median(target[self.window].data) / target.pixel_area
                )

    def iter_nm(self, n):
        nm = []
        for n_i in range(n + 1):
            for m_i in range(-n_i, n_i + 1, 2):
                nm.append((n_i, m_i))
        return nm

    @staticmethod
    @lru_cache(maxsize=1024)
    def coefficients(n, m):
        C = []
        for k in range(int((n - abs(m)) / 2) + 1):
            C.append(
                (
                    k,
                    (-1) ** k * binom(n - k, k) * binom(n - 2 * k, (n - abs(m)) / 2 - k),
                )
            )
        return C

    def Z_n_m(self, rho, phi, n, m, efficient=True):
        Z = torch.zeros_like(rho)
        if efficient:
            T_cache = {0: None}
            R_cache = {}
        for k, c in self.coefficients(n, m):
            if efficient:
                if (n - 2 * k) not in R_cache:
                    R_cache[n - 2 * k] = rho ** (n - 2 * k)
                R = R_cache[n - 2 * k]
                if m not in T_cache:
                    if m < 0:
                        T_cache[m] = torch.sin(abs(m) * phi)
                    elif m > 0:
                        T_cache[m] = torch.cos(m * phi)
                T = T_cache[m]
            else:
                R = rho ** (n - 2 * k)
                if m < 0:
                    T = torch.sin(abs(m) * phi)
                elif m > 0:
                    T = torch.cos(m * phi)

            if m == 0:
                Z += c * R
            elif m < 0:
                Z += c * R * T
            else:
                Z += c * R * T
        return Z

    @default_internal
    def evaluate_model(self, X=None, Y=None, image=None, parameters=None):
        if X is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]

        phi = self.angular_metric(X, Y, image, parameters)

        r = self.radius_metric(X, Y, image, parameters)
        r = r / self.r_scale

        G = torch.zeros_like(X)

        i = 0
        A = image.pixel_area * parameters["Anm"].value
        for n, m in self.nm_list:
            G += A[i] * self.Z_n_m(r, phi, n, m)
            i += 1

        G[r > 1] = 0.0

        return G



import unittest

import numpy as np
import matplotlib.pyplot as plt

import astrophot as ap
from utils import make_basic_sersic, make_basic_gaussian_psf


class TestPlots(unittest.TestCase):
    """
    Can't test visuals, so this only tests that the code runs
    """

    def test_target_image(self):
        target = make_basic_sersic()

        try:
            fig, ax = plt.subplots()
        except Exception:
            print("skipping test_target_image because matplotlib is not installed properly")
            return

        ap.plots.target_image(fig, ax, target)
        plt.close()

    def test_psf_image(self):
        target = make_basic_gaussian_psf()

        fig, ax = plt.subplots()

        ap.plots.psf_image(fig, ax, target)
        plt.close()

    def test_target_image_list(self):
        target1 = make_basic_sersic()
        target2 = make_basic_sersic()
        target = ap.image.Target_Image_List([target1, target2])

        try:
            fig, ax = plt.subplots(2)
        except Exception:
            print("skipping test_target_image_list because matplotlib is not installed properly")
            return

        ap.plots.target_image(fig, ax, target)
        plt.close()

    def test_model_image(self):
        target = make_basic_sersic()

        new_model = ap.models.AstroPhot_Model(
            name="constrained sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
        )
        new_model.initialize()

        try:
            fig, ax = plt.subplots()
        except Exception:
            print("skipping test because matplotlib is not installed properly")
            return

        ap.plots.model_image(fig, ax, new_model)

        plt.close()

    def test_residual_image(self):
        target = make_basic_sersic()

        new_model = ap.models.AstroPhot_Model(
            name="constrained sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
        )
        new_model.initialize()

        try:
            fig, ax = plt.subplots()
        except Exception:
            print("skipping test because matplotlib is not installed properly")
            return

        ap.plots.residual_image(fig, ax, new_model)

        plt.close()

    def test_model_windows(self):

        target = make_basic_sersic()

        new_model = ap.models.AstroPhot_Model(
            name="constrained sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
        )
        new_model.initialize()

        try:
            fig, ax = plt.subplots()
        except Exception:
            print("skipping test because matplotlib is not installed properly")
            return

        ap.plots.model_window(fig, ax, new_model)

        plt.close()



import unittest
import astrophot as ap
import numpy as np
import torch


######################################################################
# Window List Object
######################################################################


class TestWindowList(unittest.TestCase):
    def test_windowlist_creation(self):

        window1 = ap.image.Window(origin=(0, 6), pixel_shape=(100, 110))
        window2 = ap.image.Window(origin=(0, 6), pixel_shape=(100, 110))
        windowlist = ap.image.Window_List([window1, window2])

        windowlist.to(dtype=torch.float64, device="cpu")

        # under review
        self.assertEqual(windowlist.origin[0][0], 0, "Window list should capture origin")
        self.assertEqual(windowlist.origin[1][1], 6, "Window list should capture origin")
        self.assertEqual(windowlist.shape[0][0], 100, "Window list should capture shape")
        self.assertEqual(windowlist.shape[1][1], 110, "Window list should capture shape")
        self.assertEqual(windowlist.center[1][0], 50.0, "Window should determine center")
        self.assertEqual(windowlist.center[0][1], 61.0, "Window should determine center")

        x = str(windowlist)
        x = repr(windowlist)

    def test_window_arithmetic(self):

        windowbig = ap.image.Window(origin=(0, 0), pixel_shape=(100, 110))
        windowsmall = ap.image.Window(origin=(40, 40), pixel_shape=(20, 30))
        windowlistbs = ap.image.Window_List([windowbig, windowsmall])
        windowlistbb = ap.image.Window_List([windowbig, windowbig])
        windowlistsb = ap.image.Window_List([windowsmall, windowbig])

        # Logical or, size
        ######################################################################
        big_or_small = windowlistbs | windowlistsb

        self.assertEqual(
            big_or_small.origin[0][0],
            0.0,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_small.origin[1][0],
            0.0,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_small.shape[0][0],
            100,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowsmall.origin[0],
            40,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowsmall.shape[0],
            20,
            "logical or of images should not affect initial images",
        )

        # Logical and, size
        ######################################################################
        big_and_small = windowlistbs & windowlistsb
        self.assertEqual(
            big_and_small.origin[0][0],
            40,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_small.shape[0][0],
            20,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_small.shape[0][1],
            30,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowsmall.origin[0],
            40,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowsmall.shape[0],
            20,
            "logical and of images should not affect initial images",
        )

        # Logical or, offset
        ######################################################################
        windowoffset = ap.image.Window(origin=(40, -20), pixel_shape=(100, 90))
        windowlistoffset = ap.image.Window_List([windowoffset, windowoffset])
        big_or_offset = windowlistbb | windowlistoffset
        self.assertEqual(
            big_or_offset.origin[0][0],
            0,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_offset.origin[1][1],
            -20,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_offset.shape[0][0],
            140,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_offset.shape[1][1],
            130,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowoffset.origin[0],
            40,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowoffset.shape[0],
            100,
            "logical or of images should not affect initial images",
        )

        # Logical and,  offset
        ######################################################################
        big_and_offset = windowlistbb & windowlistoffset
        self.assertEqual(
            big_and_offset.origin[0][0],
            40,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_offset.origin[0][1],
            0,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_offset.shape[0][0],
            60,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_offset.shape[0][1],
            70,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowoffset.origin[0],
            40,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowoffset.shape[0],
            100,
            "logical and of images should not affect initial images",
        )

    def test_windowlist_logic(self):

        window1 = ap.image.Window(origin=[0.0, 1.0], pixel_shape=[10.2, 11.8])
        window2 = ap.image.Window(origin=[0.0, 1.0], pixel_shape=[10.2, 11.8])
        window3 = ap.image.Window(origin=[-0.6, 0.4], pixel_shape=[15.2, 18.0])
        windowlist1 = ap.image.Window_List([window1, window1.copy()])
        windowlist2 = ap.image.Window_List([window2, window2.copy()])
        windowlist3 = ap.image.Window_List([window3, window3.copy()])

        self.assertEqual(
            windowlist1, windowlist2, "same origin, shape windows should evaluate equal"
        )
        self.assertNotEqual(windowlist1, windowlist3, "Different windows should not evaluate equal")

    def test_image_list_errors(self):
        window1 = ap.image.Window(origin=[0.0, 1.0], pixel_shape=[10.2, 11.8])
        window2 = ap.image.Window(origin=[0.0, 1.0], pixel_shape=[10.2, 11.8])
        windowlist1 = ap.image.Window_List([window1, window2])

        # Bad ra dec reference point
        window2 = ap.image.Window(
            origin=[0.0, 1.0], reference_radec=np.ones(2), pixel_shape=[10.2, 11.8]
        )
        with self.assertRaises(ap.errors.ConflicingWCS):
            test_image = ap.image.Window_List((window1, window2))

        # Bad tangent plane x y reference point
        window2 = ap.image.Window(
            origin=[0.0, 1.0], reference_planexy=np.ones(2), pixel_shape=[10.2, 11.8]
        )
        with self.assertRaises(ap.errors.ConflicingWCS):
            test_image = ap.image.Window_List((window1, window2))

        # Bad WCS projection
        window2 = ap.image.Window(
            origin=[0.0, 1.0], projection="orthographic", pixel_shape=[10.2, 11.8]
        )
        with self.assertRaises(ap.errors.ConflicingWCS):
            test_image = ap.image.Window_List((window1, window2))


if __name__ == "__main__":
    unittest.main()



from .group_model_object import Group_Model
from ..image import PSF_Image
from ..errors import InvalidTarget

__all__ = ["PSF_Group_Model"]


class PSF_Group_Model(Group_Model):

    model_type = f"psf {Group_Model.model_type}"
    usable = True

    @property
    def psf_mode(self):
        return "none"

    @psf_mode.setter
    def psf_mode(self, value):
        pass

    @property
    def target(self):
        try:
            return self._target
        except AttributeError:
            return None

    @target.setter
    def target(self, tar):
        if not (tar is None or isinstance(tar, PSF_Image)):
            raise InvalidTarget("Group_Model target must be a PSF_Image instance.")
        self._target = tar

        if hasattr(self, "models"):
            for model in self.models.values():
                model.target = tar



from typing import Optional, Sequence
from collections import OrderedDict

import torch

from .core_model import AstroPhot_Model
from ..image import (
    Image,
    Target_Image,
    Image_List,
    Window,
    Window_List,
    Jacobian_Image,
)
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ._shared_methods import select_target
from ..param import Parameter_Node
from ..errors import InvalidTarget

__all__ = ["Group_Model"]


class Group_Model(AstroPhot_Model):
    """Model object which represents a list of other models. For each
    general AstroPhot model method, this calls all the appropriate
    models from its list and combines their output into a single
    summed model. This class should be used when describing any
    system more complex than makes sense to represent with a single
    light distribution.

    Args:
        name (str): unique name for the full group model
        target (Target_Image): the target image that this group model is trying to fit to
        models (Optional[Sequence[AstroPhot_Model]]): list of AstroPhot_Model objects which will combine for the group model
        locked (bool): if the whole group of models should be locked

    """

    model_type = f"group {AstroPhot_Model.model_type}"
    usable = True

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        models: Optional[Sequence[AstroPhot_Model]] = None,
        **kwargs,
    ):
        super().__init__(name=name, models=models, **kwargs)
        self._param_tuple = None
        self.models = OrderedDict()
        if models is not None:
            self.add_model(models)
        self._psf_mode = "none"
        self.update_window()
        if "filename" in kwargs:
            self.load(kwargs["filename"], new_name=name)

    def add_model(self, model):
        """Adds a new model to the group model list. Ensures that the same
        model isn't added a second time.

        Parameters:
            model: a model object to add to the model list.

        """
        if isinstance(model, (tuple, list)):
            for mod in model:
                self.add_model(mod)
            return
        if model.name in self.models and model is not self.models[model.name]:
            raise KeyError(
                f"{self.name} already has model with name {model.name}, every model must have a unique name."
            )

        self.models[model.name] = model
        self.parameters.link(model.parameters)
        self.update_window()

    def update_window(self, include_locked: bool = False):
        """Makes a new window object which encloses all the windows of the
        sub models in this group model object.

        """
        if isinstance(self.target, Image_List):  # Window_List if target is a Target_Image_List
            new_window = [None] * len(self.target.image_list)
            for model in self.models.values():
                if model.locked and not include_locked:
                    continue
                if isinstance(model.target, Image_List):
                    for target, window in zip(model.target, model.window):
                        index = self.target.index(target)
                        if new_window[index] is None:
                            new_window[index] = window.copy()
                        else:
                            new_window[index] |= window
                elif isinstance(model.target, Target_Image):
                    index = self.target.index(model.target)
                    if new_window[index] is None:
                        new_window[index] = model.window.copy()
                    else:
                        new_window[index] |= model.window
                else:
                    raise NotImplementedError(
                        f"Group_Model cannot construct a window for itself using {type(model.target)} object. Must be a Target_Image"
                    )
            new_window = Window_List(new_window)
        else:
            new_window = None
            for model in self.models.values():
                if model.locked and not include_locked:
                    continue
                if new_window is None:
                    new_window = model.window.copy()
                else:
                    new_window |= model.window
        self.window = new_window

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target: Optional[Image] = None, parameters=None, **kwargs):
        """
        Initialize each model in this group. Does this by iteratively initializing a model then subtracting it from a copy of the target.

        Args:
          target (Optional["Target_Image"]): A Target_Image instance to use as the source for initializing the model parameters on this image.
        """
        self._param_tuple = None
        super().initialize(target=target, parameters=parameters)

        target_copy = target.copy()
        for model in self.models.values():
            model.initialize(target=target_copy, parameters=parameters[model.name])
            target_copy -= model(parameters=parameters[model.name])

    def fit_mask(self) -> torch.Tensor:
        """Returns a mask for the target image which is the combination of all
        the fit masks of the sub models. This mask is used when the multiple
        models in the group model do not completely overlap with each other, thus
        there are some pixels which are not covered by any model and have no
        reason to be fit.

        """
        if isinstance(self.target, Image_List):
            mask = tuple(torch.ones_like(submask) for submask in self.target[self.window].mask)
            for model in self.models.values():
                model_flat_mask = model.fit_mask()
                if isinstance(model.target, Image_List):
                    for target, window, submask in zip(model.target, model.window, model_flat_mask):
                        index = self.target.index(target)
                        group_indices = self.window.window_list[index].get_self_indices(window)
                        model_indices = window.get_self_indices(self.window.window_list[index])
                        mask[index][group_indices] &= submask[model_indices]
                else:
                    index = self.target.index(model.target)
                    group_indices = self.window.window_list[index].get_self_indices(model.window)
                    model_indices = model.window.get_self_indices(self.window.window_list[index])
                    mask[index][group_indices] &= model_flat_mask[model_indices]
        else:
            mask = torch.ones_like(self.target[self.window].mask)
            for model in self.models.values():
                group_indices = self.window.get_self_indices(model.window)
                model_indices = model.window.get_self_indices(self.window)
                mask[group_indices] &= model.fit_mask()[model_indices]
        return mask

    def sample(
        self,
        image: Optional[Image] = None,
        window: Optional[Window] = None,
        parameters: Optional["Parameter_Node"] = None,
    ):
        """Sample the group model on an image. Produces the flux values for
        each pixel associated with the models in this group. Each
        model is called individually and the results are added
        together in one larger image.

        Args:
          image (Optional["Model_Image"]): Image to sample on, overrides the windows for each sub model, they will all be evaluated over this entire image. If left as none then each sub model will be evaluated in its window.

        """
        self._param_tuple = None
        if image is None:
            sample_window = True
            image = self.make_model_image(window=window)
        else:
            sample_window = False
        if parameters is None:
            parameters = self.parameters

        for model in self.models.values():
            if window is not None and isinstance(window, Window_List):
                indices = self.target.match_indices(model.target)
                if isinstance(indices, (tuple, list)):
                    use_window = Window_List(
                        window_list=list(window.window_list[ind] for ind in indices)
                    )
                else:
                    use_window = window.window_list[indices]
            else:
                use_window = window
            if sample_window:
                # Will sample the model fit window then add to the image
                image += model(window=use_window, parameters=parameters[model.name])
            else:
                # Will sample the entire image
                model(image, window=use_window, parameters=parameters[model.name])

        return image

    @torch.no_grad()
    def jacobian(
        self,
        parameters: Optional[torch.Tensor] = None,
        as_representation: bool = False,
        pass_jacobian: Optional[Jacobian_Image] = None,
        window: Optional[Window] = None,
        **kwargs,
    ):
        """Compute the jacobian for this model. Done by first constructing a
        full jacobian (Npixels * Nparameters) of zeros then call the
        jacobian method of each sub model and add it in to the total.

        Args:
          parameters (Optional[torch.Tensor]): 1D parameter vector to overwrite current values
          as_representation (bool): Indicates if the "parameters" argument is in the form of the real values, or as representations in the (-inf,inf) range. Default False
          pass_jacobian (Optional["Jacobian_Image"]): A Jacobian image pre-constructed to be passed along instead of constructing new Jacobians

        """
        if window is None:
            window = self.window
        self._param_tuple = None

        if parameters is not None:
            if as_representation:
                self.parameters.vector_set_representation(parameters)
            else:
                self.parameters.vector_set_values(parameters)

        if pass_jacobian is None:
            jac_img = self.target[window].jacobian_image(
                parameters=self.parameters.vector_identities()
            )
        else:
            jac_img = pass_jacobian

        for model in self.models.values():
            if isinstance(model, Group_Model):
                model.jacobian(
                    as_representation=as_representation,
                    pass_jacobian=jac_img,
                    window=window,
                )
            else:  # fixme, maybe make pass_jacobian be filled internally to each model
                jac_img += model.jacobian(
                    as_representation=as_representation,
                    pass_jacobian=jac_img,
                    window=window,
                )

        return jac_img

    def __iter__(self):
        return (mod for mod in self.models.values())

    @property
    def psf_mode(self):
        return self._psf_mode

    @psf_mode.setter
    def psf_mode(self, value):
        self._psf_mode = value
        for model in self.models.values():
            model.psf_mode = value

    @property
    def target(self):
        try:
            return self._target
        except AttributeError:
            return None

    @target.setter
    def target(self, tar):
        if not (tar is None or isinstance(tar, Target_Image)):
            raise InvalidTarget("Group_Model target must be a Target_Image instance.")
        self._target = tar

        if hasattr(self, "models"):
            for model in self.models.values():
                model.target = tar

    def get_state(self, save_params=True):
        """Returns a dictionary with information about the state of the model
        and its parameters.

        """
        state = super().get_state(save_params=save_params)
        if save_params:
            state["parameters"] = self.parameters.get_state()
        if "models" not in state:
            state["models"] = {}
        for model in self.models.values():
            state["models"][model.name] = model.get_state(save_params=False)
        return state

    def load(self, filename="AstroPhot.yaml", new_name=None):
        """Loads an AstroPhot state file and updates this model with the
        loaded parameters.

        """
        state = AstroPhot_Model.load(filename)

        if new_name is None:
            new_name = state["name"]
        self.name = new_name

        if isinstance(state["parameters"], Parameter_Node):
            self.parameters = state["parameters"]
        else:
            self.parameters = Parameter_Node(self.name, state=state["parameters"])

        for model in state["models"]:
            state["models"][model]["parameters"] = self.parameters[model]
            for own_model in self.models.values():
                if model == own_model.name:
                    own_model.load(state["models"][model])
                    break
            else:
                self.add_model(
                    AstroPhot_Model(name=model, filename=state["models"][model], target=self.target)
                )
        self.update_window()



import unittest
import astrophot as ap
import numpy as np
import torch


class TestWPCS(unittest.TestCase):
    def test_wpcs_creation(self):

        # Blank startup
        wcs_blank = ap.image.WPCS()

        self.assertEqual(wcs_blank.projection, "gnomonic", "Default projection should be Gnomonic")
        self.assertTrue(
            torch.all(wcs_blank.reference_radec == 0),
            "default reference world coordinates should be zeros",
        )
        self.assertTrue(
            torch.all(wcs_blank.reference_planexy == 0),
            "default reference plane coordinates should be zeros",
        )

        # Provided parameters
        wcs_set = ap.image.WPCS(
            projection="orthographic",
            reference_radec=(90, 10),
        )

        self.assertEqual(wcs_set.projection, "orthographic", "Provided projection was Orthographic")
        self.assertTrue(
            torch.all(
                wcs_set.reference_radec
                == torch.tensor(
                    (90, 10), dtype=ap.AP_config.ap_dtype, device=ap.AP_config.ap_device
                )
            ),
            "World coordinates should be as provided",
        )
        self.assertNotEqual(
            wcs_blank.projection,
            "orthographic",
            "Not all WCS objects should be updated",
        )
        self.assertFalse(
            torch.all(
                wcs_blank.reference_radec
                == torch.tensor(
                    (90, 10), dtype=ap.AP_config.ap_dtype, device=ap.AP_config.ap_device
                )
            ),
            "Not all WCS objects should be updated",
        )

        wcs_set = wcs_set.copy()

        self.assertEqual(wcs_set.projection, "orthographic", "Provided projection was Orthographic")
        self.assertTrue(
            torch.all(
                wcs_set.reference_radec
                == torch.tensor(
                    (90, 10), dtype=ap.AP_config.ap_dtype, device=ap.AP_config.ap_device
                )
            ),
            "World coordinates should be as provided",
        )
        self.assertNotEqual(
            wcs_blank.projection,
            "orthographic",
            "Not all WCS objects should be updated",
        )
        self.assertFalse(
            torch.all(
                wcs_blank.reference_radec
                == torch.tensor(
                    (90, 10), dtype=ap.AP_config.ap_dtype, device=ap.AP_config.ap_device
                )
            ),
            "Not all WCS objects should be updated",
        )

    def test_wpcs_round_trip(self):

        for projection in ["gnomonic", "orthographic", "steriographic"]:
            print(projection)
            for ref_coords in [(20.3, 79), (120.2, -19), (300, -50), (0, 0)]:
                print(ref_coords)
                wcs = ap.image.WPCS(
                    projection=projection,
                    reference_radec=ref_coords,
                )

                test_grid_RA, test_grid_DEC = torch.meshgrid(
                    torch.linspace(
                        ref_coords[0] - 10,
                        ref_coords[0] + 10,
                        10,
                        dtype=ap.AP_config.ap_dtype,
                        device=ap.AP_config.ap_device,
                    ),  # RA
                    torch.linspace(
                        ref_coords[1] - 10,
                        ref_coords[1] + 10,
                        10,
                        dtype=ap.AP_config.ap_dtype,
                        device=ap.AP_config.ap_device,
                    ),  # DEC
                    indexing="xy",
                )

                project_x, project_y = wcs.world_to_plane(
                    test_grid_RA,
                    test_grid_DEC,
                )

                reproject_RA, reproject_DEC = wcs.plane_to_world(
                    project_x,
                    project_y,
                )

                self.assertTrue(
                    torch.allclose(reproject_RA, test_grid_RA),
                    "Round trip RA should map back to itself",
                )
                self.assertTrue(
                    torch.allclose(reproject_DEC, test_grid_DEC),
                    "Round trip DEC should map back to itself",
                )

    def test_wpcs_errors(self):
        with self.assertRaises(ap.errors.InvalidWCS):
            wcs = ap.image.WPCS(
                projection="connor",
            )


class TestPPCS(unittest.TestCase):

    def test_ppcs_creation(self):
        # Blank startup
        wcs_blank = ap.image.PPCS()

        self.assertTrue(
            np.all(
                wcs_blank.pixelscale.detach().cpu().numpy() == np.array([[1.0, 0.0], [0.0, 1.0]])
            ),
            "Default pixelscale should be 1",
        )
        self.assertTrue(
            torch.all(wcs_blank.reference_imageij == -0.5),
            "default reference pixel coordinates should be -0.5",
        )
        self.assertTrue(
            torch.all(wcs_blank.reference_imagexy == 0.0),
            "default reference plane coordinates should be zeros",
        )

        # Provided parameters
        wcs_set = ap.image.PPCS(
            pixelscale=[[-0.173205, 0.1], [0.15, 0.259808]],
            reference_imageij=(5, 10),
            reference_imagexy=(0.12, 0.45),
        )

        self.assertTrue(
            torch.allclose(
                wcs_set.pixelscale,
                torch.tensor(
                    [[-0.173205, 0.1], [0.15, 0.259808]],
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "Provided pixelscale should be used",
        )
        self.assertTrue(
            torch.allclose(
                wcs_set.reference_imageij,
                torch.tensor(
                    (5.0, 10.0),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "pixel reference coordinates should be as provided",
        )
        self.assertTrue(
            torch.allclose(
                wcs_set.reference_imagexy,
                torch.tensor(
                    (0.12, 0.45),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "plane reference coordinates should be as provided",
        )
        self.assertTrue(
            torch.allclose(
                wcs_set.plane_to_pixel(
                    torch.tensor(
                        (0.12, 0.45),
                        dtype=ap.AP_config.ap_dtype,
                        device=ap.AP_config.ap_device,
                    )
                ),
                torch.tensor(
                    (5.0, 10.0),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "plane reference coordinates should map to pixel reference coordinates",
        )
        self.assertTrue(
            torch.allclose(
                wcs_set.pixel_to_plane(
                    torch.tensor(
                        (5.0, 10.0),
                        dtype=ap.AP_config.ap_dtype,
                        device=ap.AP_config.ap_device,
                    )
                ),
                torch.tensor(
                    (0.12, 0.45),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "pixel reference coordinates should map to plane reference coordinates",
        )

        wcs_set = wcs_set.copy()

        self.assertTrue(
            torch.allclose(
                wcs_set.pixelscale,
                torch.tensor(
                    [[-0.173205, 0.1], [0.15, 0.259808]],
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "Provided pixelscale should be used",
        )
        self.assertTrue(
            torch.allclose(
                wcs_set.reference_imageij,
                torch.tensor(
                    (5.0, 10.0),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "pixel reference coordinates should be as provided",
        )
        self.assertTrue(
            torch.allclose(
                wcs_set.reference_imagexy,
                torch.tensor(
                    (0.12, 0.45),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "plane reference coordinates should be as provided",
        )
        self.assertTrue(
            torch.allclose(
                wcs_set.plane_to_pixel(
                    torch.tensor(
                        (0.12, 0.45),
                        dtype=ap.AP_config.ap_dtype,
                        device=ap.AP_config.ap_device,
                    )
                ),
                torch.tensor(
                    (5.0, 10.0),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "plane reference coordinates should map to pixel reference coordinates",
        )
        self.assertTrue(
            torch.allclose(
                wcs_set.pixel_to_plane(
                    torch.tensor(
                        (5.0, 10.0),
                        dtype=ap.AP_config.ap_dtype,
                        device=ap.AP_config.ap_device,
                    )
                ),
                torch.tensor(
                    (0.12, 0.45),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "pixel reference coordinates should map to plane reference coordinates",
        )

        wcs_set.pixelscale = None

    def test_ppcs_round_trip(self):

        for pixelscale in [
            0.2,
            [[0.6, 0.0], [0.0, 0.4]],
            [[-0.173205, 0.1], [0.15, 0.259808]],
        ]:
            print(pixelscale)
            for ref_coords in [(20.3, 79), (120.2, -19), (300, -50), (0, 0)]:
                print(ref_coords)
                wcs = ap.image.PPCS(
                    pixelscale=pixelscale,
                    reference_imagexy=ref_coords,
                )

                test_grid_x, test_grid_y = torch.meshgrid(
                    torch.linspace(
                        ref_coords[0] - 10,
                        ref_coords[0] + 10,
                        10,
                        dtype=ap.AP_config.ap_dtype,
                        device=ap.AP_config.ap_device,
                    ),  # x
                    torch.linspace(
                        ref_coords[1] - 10,
                        ref_coords[1] + 10,
                        10,
                        dtype=ap.AP_config.ap_dtype,
                        device=ap.AP_config.ap_device,
                    ),  # y
                    indexing="xy",
                )

                project_i, project_j = wcs.plane_to_pixel(
                    test_grid_x,
                    test_grid_y,
                )

                reproject_x, reproject_y = wcs.pixel_to_plane(
                    project_i,
                    project_j,
                )

                self.assertTrue(
                    torch.allclose(reproject_x, test_grid_x),
                    "Round trip x should map back to itself",
                )
                self.assertTrue(
                    torch.allclose(reproject_y, test_grid_y),
                    "Round trip y should map back to itself",
                )


class TestWCS(unittest.TestCase):
    def test_wcs_creation(self):

        wcs = ap.image.WCS(
            projection="orthographic",
            pixelscale=[[-0.173205, 0.1], [0.15, 0.259808]],
            reference_radec=(120.2, -19),
            reference_imagexy=(33.0, 123.0),
        )

        wcs2 = wcs.copy()

        self.assertEqual(wcs2.projection, "orthographic", "Provided projection was Orthographic")
        self.assertTrue(
            torch.allclose(wcs2.reference_radec, wcs.reference_radec),
            "World coordinates should be as provided",
        )
        self.assertTrue(
            torch.allclose(wcs2.reference_planexy, wcs.reference_planexy),
            "Plane coordinates should be as provided",
        )
        self.assertTrue(
            torch.allclose(wcs2.reference_imagexy, wcs.reference_imagexy),
            "imagexy coordinates should be as provided",
        )
        self.assertTrue(
            torch.allclose(wcs2.reference_imageij, wcs.reference_imageij),
            "imageij coordinates should be as provided",
        )
        self.assertTrue(
            torch.allclose(wcs2.pixelscale, wcs.pixelscale),
            "pixelscale should be as provided",
        )

    def test_wcs_roundtrip(self):
        for pixelscale in [
            0.2,
            [[0.6, 0.0], [0.0, 0.4]],
            [[-0.173205, 0.1], [0.15, 0.259808]],
        ]:
            print(pixelscale)
            for ref_coords_xy in [(33.0, 123.0), (-430.2, -11), (-97.0, 5), (0, 0)]:
                for projection in ["gnomonic", "orthographic", "steriographic"]:
                    print(projection)
                    for ref_coords_radec in [
                        (20.3, 79),
                        (120.2, -19),
                        (300, -50),
                        (0, 0),
                    ]:
                        print(ref_coords_radec)
                        wcs = ap.image.WCS(
                            projection=projection,
                            pixelscale=pixelscale,
                            reference_radec=ref_coords_radec,
                            reference_imagexy=ref_coords_xy,
                        )

                        test_grid_RA, test_grid_DEC = torch.meshgrid(
                            torch.linspace(
                                ref_coords_radec[0] - 10,
                                ref_coords_radec[0] + 10,
                                10,
                                dtype=ap.AP_config.ap_dtype,
                                device=ap.AP_config.ap_device,
                            ),  # RA
                            torch.linspace(
                                ref_coords_radec[1] - 10,
                                ref_coords_radec[1] + 10,
                                10,
                                dtype=ap.AP_config.ap_dtype,
                                device=ap.AP_config.ap_device,
                            ),  # DEC
                            indexing="xy",
                        )

                        project_i, project_j = wcs.world_to_pixel(
                            test_grid_RA,
                            test_grid_DEC,
                        )

                        reproject_RA, reproject_DEC = wcs.pixel_to_world(
                            project_i,
                            project_j,
                        )

                        self.assertTrue(
                            torch.allclose(reproject_RA, test_grid_RA),
                            "Round trip RA should map back to itself",
                        )
                        self.assertTrue(
                            torch.allclose(reproject_DEC, test_grid_DEC),
                            "Round trip DEC should map back to itself",
                        )

    def test_wcs_state(self):
        wcs = ap.image.WCS(
            projection="orthographic",
            pixelscale=[[-0.173205, 0.1], [0.15, 0.259808]],
            reference_radec=(120.2, -19),
            reference_imagexy=(33.0, 123.0),
        )

        wcs_state = wcs.get_state()

        new_wcs = ap.image.WCS(state=wcs_state)

        self.assertEqual(
            wcs.projection, new_wcs.projection, "WCS projection should be set by state"
        )
        self.assertTrue(
            torch.allclose(
                wcs.pixelscale,
                torch.tensor(
                    [[-0.173205, 0.1], [0.15, 0.259808]],
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "WCS pixelscale should be set by state",
        )
        self.assertTrue(
            torch.allclose(
                wcs.reference_radec,
                torch.tensor(
                    (120.2, -19),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "WCS reference RA DEC should be set by state",
        )
        self.assertTrue(
            torch.allclose(
                wcs.reference_imagexy,
                torch.tensor(
                    (33.0, 123.0),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "WCS reference image position should be set by state",
        )

        wcs_state = wcs.get_fits_state()

        new_wcs = ap.image.WCS()
        new_wcs.set_fits_state(state=wcs_state)

        self.assertEqual(
            wcs.projection, new_wcs.projection, "WCS projection should be set by state"
        )
        self.assertTrue(
            torch.allclose(
                wcs.pixelscale,
                torch.tensor(
                    [[-0.173205, 0.1], [0.15, 0.259808]],
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "WCS pixelscale should be set by state",
        )
        self.assertTrue(
            torch.allclose(
                wcs.reference_radec,
                torch.tensor(
                    (120.2, -19),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "WCS reference RA DEC should be set by state",
        )
        self.assertTrue(
            torch.allclose(
                wcs.reference_imagexy,
                torch.tensor(
                    (33.0, 123.0),
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
            ),
            "WCS reference image position should be set by state",
        )

    def test_wcs_repr(self):

        wcs = ap.image.WCS(
            projection="orthographic",
            pixelscale=[[-0.173205, 0.1], [0.15, 0.259808]],
            reference_radec=(120.2, -19),
            reference_imagexy=(33.0, 123.0),
        )

        S = str(wcs)
        R = repr(wcs)



# Apply an optimizer toa  downsampled version of an image
from typing import Dict, Any

import numpy as np

from .base import BaseOptimizer
from ..models import AstroPhot_Model
from .lm import LM
from .. import AP_config

__all__ = ["MiniFit"]


class MiniFit(BaseOptimizer):
    def __init__(
        self,
        model: AstroPhot_Model,
        downsample_factor: int = 1,
        max_pixels: int = 10000,
        method: BaseOptimizer = LM,
        initial_state: np.ndarray = None,
        method_kwargs: Dict[str, Any] = {},
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(model, initial_state, **kwargs)

        self.method = method
        self.method_kwargs = method_kwargs
        if "verbose" not in self.method_kwargs:
            self.method_kwargs["verbose"] = self.verbose

        self.downsample_factor = downsample_factor
        self.max_pixels = max_pixels

    def fit(self) -> BaseOptimizer:
        initial_target = self.model.target
        target_area = self.model.target[self.model.window]
        while True:
            small_target = target_area.reduce(self.downsample_factor)
            if small_target.size < self.max_pixels:
                break
            self.downsample_factor += 1

        if self.verbose > 0:
            AP_config.ap_logger.info(f"Downsampling target by {self.downsample_factor}x")

        self.small_target = small_target
        self.model.target = small_target
        res = self.method(self.model, **self.method_kwargs).fit()
        self.model.target = initial_target

        self.message = res.message

        return self



from functools import partial

import numpy as np
import torch
from scipy.stats import binned_statistic, iqr

from .. import AP_config
from ..models import Warp_Galaxy
from ..utils.conversions.units import flux_to_sb
from .visuals import *
from ..errors import InvalidModel

__all__ = [
    "radial_light_profile",
    "radial_median_profile",
    "ray_light_profile",
    "wedge_light_profile",
    "warp_phase_profile",
]


def radial_light_profile(
    fig,
    ax,
    model,
    rad_unit="arcsec",
    extend_profile=1.0,
    R0=0.0,
    resolution=1000,
    doassert=True,
    plot_kwargs={},
):
    xx = torch.linspace(
        R0,
        torch.max(model.window.shape / 2) * extend_profile,
        int(resolution),
        dtype=AP_config.ap_dtype,
        device=AP_config.ap_device,
    )
    flux = model.radial_model(xx).detach().cpu().numpy()
    if model.target.zeropoint is not None:
        yy = flux_to_sb(flux, model.target.pixel_area.item(), model.target.zeropoint.item())
    else:
        yy = np.log10(flux)

    kwargs = {
        "linewidth": 2,
        "color": main_pallet["primary1"],
        "label": f"{model.name} profile",
    }
    kwargs.update(plot_kwargs)
    with torch.no_grad():
        ax.plot(
            xx.detach().cpu().numpy(),
            yy,
            **kwargs,
        )

    if model.target.zeropoint is not None:
        ax.set_ylabel("Surface Brightness [mag/arcsec$^2$]")
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
    else:
        ax.set_ylabel("log$_{10}$(flux/arcsec$^2$)")
    ax.set_xlabel(f"Radius [{rad_unit}]")
    ax.set_xlim([R0, None])
    return fig, ax


def radial_median_profile(
    fig,
    ax,
    model: "AstroPhot_Model",
    count_limit: int = 10,
    return_profile: bool = False,
    rad_unit: str = "arcsec",
    doassert: bool = True,
    plot_kwargs: dict = {},
):
    """Plot an SB profile by taking flux median at each radius.

    Using the coordinate transforms defined by the model object,
    assigns a radius to each pixel then bins the pixel-radii and
    computes the median in each bin. This gives a simple
    representation of the image data if one were to simply average the
    pixels along isophotes.

    Args:
      fig: matplotlib figure object
      ax: matplotlib axis object
      model (AstroPhot_Model): Model object from which to determine the radial binning. Also provides the target image to extract the data
      count_limit (int): The limit of pixels in a bin, below which uncertainties are not computed. Default: 10
      return_profile (bool): Instead of just returning the fig and ax object, will return the extracted profile formatted as: Rbins (the radial bin edges), medians (the median in each bin), scatter (the 16-84 quartile range / 2), count (the number of pixels in each bin). Default: False
      rad_unit (str): The name of the physical radius units. Default: "arcsec"
      doassert (bool): If any requirements are imposed on which kind of profile can be plotted, this activates them. Default: True

    """

    Rlast_phys = torch.max(model.window.shape / 2).item()
    Rlast_pix = Rlast_phys / model.target.pixel_length.item()

    Rbins = [0.0]
    while Rbins[-1] < Rlast_pix:
        Rbins.append(Rbins[-1] + max(2, Rbins[-1] * 0.1))
    Rbins = np.array(Rbins)

    with torch.no_grad():
        image = model.target[model.window]
        X, Y = image.get_coordinate_meshgrid() - model["center"].value[..., None, None]
        X, Y = model.transform_coordinates(X, Y)
        R = model.radius_metric(X, Y)
        R = R.detach().cpu().numpy()

    count, bins, binnum = binned_statistic(
        R.ravel(),
        image.data.detach().cpu().numpy().ravel(),
        statistic="count",
        bins=Rbins,
    )

    stat, bins, binnum = binned_statistic(
        R.ravel(),
        image.data.detach().cpu().numpy().ravel(),
        statistic="median",
        bins=Rbins,
    )
    stat[count == 0] = np.nan

    scat, bins, binnum = binned_statistic(
        R.ravel(),
        image.data.detach().cpu().numpy().ravel(),
        statistic=partial(iqr, rng=(16, 84)),
        bins=Rbins,
    )
    scat[count > count_limit] /= 2 * np.sqrt(count[count > count_limit])
    scat[count <= count_limit] = 0

    if model.target.zeropoint is not None:
        stat = flux_to_sb(stat, model.target.pixel_area.item(), model.target.zeropoint.item())
        ax.set_ylabel("Surface Brightness [mag/arcsec$^2$]")
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
    else:
        stat = np.log10(stat)
        ax.set_ylabel("log$_{10}$(flux/arcsec^2)")

    kwargs = {
        "linewidth": 0,
        "elinewidth": 1,
        "color": main_pallet["primary2"],
        "label": "data profile",
    }
    kwargs.update(plot_kwargs)
    ax.errorbar(
        (Rbins[:-1] + Rbins[1:]) / 2,
        stat,
        yerr=scat,
        fmt=".",
        **kwargs,
    )
    ax.set_xlabel(f"Radius [{rad_unit}]")

    if return_profile:
        return Rbins, stat, scat, count
    return fig, ax


def ray_light_profile(
    fig,
    ax,
    model,
    rad_unit="arcsec",
    extend_profile=1.0,
    resolution=1000,
    doassert=True,
):
    xx = torch.linspace(
        0,
        torch.max(model.window.shape / 2) * extend_profile,
        int(resolution),
        dtype=AP_config.ap_dtype,
        device=AP_config.ap_device,
    )
    for r in range(model.rays):
        if model.rays <= 5:
            col = main_pallet[f"primary{r+1}"]
        else:
            col = cmap_grad(r / model.rays)
        with torch.no_grad():
            ax.plot(
                xx.detach().cpu().numpy(),
                np.log10(model.iradial_model(r, xx).detach().cpu().numpy()),
                linewidth=2,
                color=col,
                label=f"{model.name} profile {r}",
            )
    ax.set_ylabel("log$_{10}$(flux)")
    ax.set_xlabel(f"Radius [{rad_unit}]")

    return fig, ax


def wedge_light_profile(
    fig,
    ax,
    model,
    rad_unit="arcsec",
    extend_profile=1.0,
    resolution=1000,
    doassert=True,
):
    xx = torch.linspace(
        0,
        torch.max(model.window.shape / 2) * extend_profile,
        int(resolution),
        dtype=AP_config.ap_dtype,
        device=AP_config.ap_device,
    )
    for r in range(model.wedges):
        if model.wedges <= 5:
            col = main_pallet[f"primary{r+1}"]
        else:
            col = cmap_grad(r / model.wedges)
        with torch.no_grad():
            ax.plot(
                xx.detach().cpu().numpy(),
                np.log10(model.iradial_model(r, xx).detach().cpu().numpy()),
                linewidth=2,
                color=col,
                label=f"{model.name} profile {r}",
            )
    ax.set_ylabel("log$_{10}$(flux)")
    ax.set_xlabel(f"Radius [{rad_unit}]")

    return fig, ax


def warp_phase_profile(fig, ax, model, rad_unit="arcsec", doassert=True):
    if doassert:
        if not isinstance(model, Warp_Galaxy):
            raise InvalidModel(
                f"warp_phase_profile must be given a 'Warp_Galaxy' object. Not {type(model)}"
            )

    ax.plot(
        model.profR,
        model["q(R)"].value.detach().cpu().numpy(),
        linewidth=2,
        color=main_pallet["primary1"],
        label=f"{model.name} axis ratio",
    )
    ax.plot(
        model.profR,
        model["PA(R)"].detach().cpu().numpy() / np.pi,
        linewidth=2,
        color=main_pallet["secondary1"],
        label=f"{model.name} position angle",
    )
    ax.set_ylim([0, 1])
    ax.set_ylabel("q [b/a], PA [rad/$\\pi$]")
    ax.set_xlabel(f"Radius [{rad_unit}]")

    return fig, ax



# Levenberg-Marquardt algorithm
from typing import Sequence
from functools import partial

import torch
import numpy as np

from .base import BaseOptimizer
from .. import AP_config
from ..errors import OptimizeStop

__all__ = ("LM",)


class LM(BaseOptimizer):
    """The LM class is an implementation of the Levenberg-Marquardt
    optimization algorithm. This method is used to solve non-linear
    least squares problems and is known for its robustness and
    efficiency.

    The Levenberg-Marquardt (LM) algorithm is an iterative method used
    for solving non-linear least squares problems. It can be seen as a
    combination of the Gauss-Newton method and the gradient descent
    method: it works like the Gauss-Newton method when the parameters
    are approximately near a minimum, and like the gradient descent method
    when the parameters are far from their optimal values.

    The cost function that the LM algorithm tries to minimize is of
    the form:

    .. math::
        f(\\boldsymbol{\\beta}) = \\frac{1}{2}\\sum_{i=1}^{N} r_i(\\boldsymbol{\\beta})^2

    where :math:`\\boldsymbol{\\beta}` is the vector of parameters,
    :math:`r_i` are the residuals, and :math:`N` is the number of
    observations.

    The LM algorithm iteratively performs the following update to the parameters:

    .. math::
        \\boldsymbol{\\beta}_{n+1} = \\boldsymbol{\\beta}_{n} - (J^T J + \\lambda diag(J^T J))^{-1} J^T \\boldsymbol{r}

    where:
        - :math:`J` is the Jacobian matrix whose elements are :math:`J_{ij} = \\frac{\\partial r_i}{\\partial \\beta_j}`,
        - :math:`\\boldsymbol{r}` is the vector of residuals :math:`r_i(\\boldsymbol{\\beta})`,
        - :math:`\\lambda` is a damping factor which is adjusted at each iteration.

    When :math:`\\lambda = 0` this can be seen as the Gauss-Newton
    method. In the limit that :math:`\\lambda` is large, the
    :math:`J^T J` matrix (an approximation of the Hessian) becomes
    subdominant and the update essentially points along :math:`J^T
    \\boldsymbol{r}` which is the gradient. In this scenario the
    gradient descent direction is also modified by the :math:`\\lambda
    diag(J^T J)` scaling which in some sense makes each gradient
    unitless and further improves the step. Note as well that as
    :math:`\\lambda` gets larger the step taken will be smaller, which
    helps to ensure convergence when the initial guess of the
    parameters are far from the optimal solution.

    Note that the residuals :math:`r_i` are typically also scaled by
    the variance of the pixels, but this does not change the equations
    above. For a detailed explanation of the LM method see the article
    by Henri Gavin on which much of the AstroPhot LM implementation is
    based::

        @article{Gavin2019,
            title={The Levenberg-Marquardt algorithm for nonlinear least squares curve-fitting problems},
            author={Gavin, Henri P},
            journal={Department of Civil and Environmental Engineering, Duke University},
            volume={19},
            year={2019}
        }

    as well as the paper on LM geodesic acceleration by Mark
    Transtrum::

        @article{Tanstrum2012,
           author = {{Transtrum}, Mark K. and {Sethna}, James P.},
            title = "{Improvements to the Levenberg-Marquardt algorithm for nonlinear least-squares minimization}",
             year = 2012,
              doi = {10.48550/arXiv.1201.5885},
           adsurl = {https://ui.adsabs.harvard.edu/abs/2012arXiv1201.5885T},
        }

    The damping factor :math:`\\lambda` is adjusted at each iteration:
    it is effectively increased when we are far from the solution, and
    decreased when we are close to it. In practice, the algorithm
    attempts to pick the smallest :math:`\\lambda` that is can while
    making sure that the :math:`\\chi^2` decreases at each step.

    The main advantage of the LM algorithm is its adaptability. When
    the current estimate is far from the optimum, the algorithm
    behaves like a gradient descent to ensure global
    convergence. However, when it is close to the optimum, it behaves
    like Gauss-Newton, which provides fast local convergence.

    In practice, the algorithm is typically implemented with various
    enhancements to improve its performance. For example, the Jacobian
    may be approximated with finite differences, geodesic acceleration
    can be used to speed up convergence, and more sophisticated
    strategies can be used to adjust the damping factor :math:`\\lambda`.

    The exact performance of the LM algorithm will depend on the
    nature of the problem, including the complexity of the function
    f(x), the quality of the initial estimate x0, and the distribution
    of the data.

    The LM class implemented for AstroPhot takes a model, initial
    state, and various other optional parameters as inputs and seeks
    to find the parameters that minimize the cost function.

    Args:
      model: The model to be optimized.
      initial_state (Sequence): Initial values for the parameters to be optimized.
      max_iter (int): Maximum number of iterations for the algorithm.
      relative_tolerance (float): Tolerance level for relative change in cost function value to trigger termination of the algorithm.
      fit_parameters_identity: Used to select a subset of parameters. This is mostly used internally.
      verbose: Controls the verbosity of the output during optimization. A higher value results in more detailed output. If not provided, defaults to 0 (no output).
      max_step_iter (optional): The maximum number of steps while searching for chi^2 improvement on a single Jacobian evaluation. Default is 10.
      curvature_limit (optional): Controls how cautious the optimizer is for changing curvature. It should be a number greater than 0, where smaller is more cautious. Default is 1.
      Lup and Ldn (optional): These adjust the step sizes for the damping parameter. Default is 5 and 3 respectively.
      L0 (optional): This is the starting damping parameter. For easy problems with good initialization, this can be set lower. Default is 1.
      acceleration (optional): Controls the use of geodesic acceleration, which can be helpful in some scenarios. Set 1 for full acceleration, 0 for no acceleration. Default is 0.

    Here is some basic usage of the LM optimizer:

    .. code-block:: python

      import astrophot as ap

      # build model
      # ...

      # Initialize model parameters
      model.initialize()

      # Fit the parameters
      result = ap.fit.lm(model, verbose=1)

      # Check that a minimum was found
      print(result.message)

      # See the minimum chi^2 value
      print(f"min chi2: {result.res_loss()}")

      # Update parameter uncertainties
      result.update_uncertainty()

      # Extract multivariate Gaussian of uncertainties
      mu = result.res()
      cov = result.covariance_matrix

    """

    def __init__(
        self,
        model,
        initial_state: Sequence = None,
        max_iter: int = 100,
        relative_tolerance: float = 1e-5,
        ndf=None,
        **kwargs,
    ):

        super().__init__(
            model,
            initial_state,
            max_iter=max_iter,
            relative_tolerance=relative_tolerance,
            **kwargs,
        )
        # The forward model which computes the output image given input parameters
        self.forward = partial(model, as_representation=True)
        # Compute the jacobian in representation units (defined for -inf, inf)
        self.jacobian = partial(model.jacobian, as_representation=True)
        self.jacobian_natural = partial(model.jacobian, as_representation=False)
        # Maximum number of iterations of the algorithm
        self.max_iter = max_iter
        # Maximum number of steps while searching for chi^2 improvement on a single jacobian evaluation
        self.max_step_iter = kwargs.get("max_step_iter", 10)
        # sets how cautious the optimizer is for changing curvature, should be number greater than 0, where smaller is more cautious
        self.curvature_limit = kwargs.get("curvature_limit", 1.0)
        # These are the adjustment step sized for the damping parameter
        self._Lup = kwargs.get("Lup", 11.0)
        self._Ldn = kwargs.get("Ldn", 9.0)
        # This is the starting damping parameter, for easy problems with good initialization, this can be set lower
        self.L = kwargs.get("L0", 1.0)
        # Geodesic acceleration is helpful in some scenarios. By default it is turned off. Set 1 for full acceleration, 0 for no acceleration.
        self.acceleration = kwargs.get("acceleration", 0.0)
        # Initialize optimizer attributes
        self.Y = self.model.target[self.fit_window].flatten("data")

        # 1 / (sigma^2)
        kW = kwargs.get("W", None)
        if kW is not None:
            self.W = torch.as_tensor(
                kW, dtype=AP_config.ap_dtype, device=AP_config.ap_device
            ).flatten()
        elif model.target.has_variance:
            self.W = self.model.target[self.fit_window].flatten("weight")
        else:
            self.W = torch.ones_like(self.Y)

        # mask
        fit_mask = self.model.fit_mask()
        if isinstance(fit_mask, tuple):
            fit_mask = torch.cat(tuple(FM.flatten() for FM in fit_mask))
        else:
            fit_mask = fit_mask.flatten()
        if torch.sum(fit_mask).item() == 0:
            fit_mask = None
        if model.target.has_mask:
            mask = self.model.target[self.fit_window].flatten("mask")
            if fit_mask is not None:
                mask = mask | fit_mask
            self.mask = torch.logical_not(mask)
        elif fit_mask is not None:
            self.mask = torch.logical_not(fit_mask)
        else:
            self.mask = None
        if self.mask is not None and torch.sum(self.mask).item() == 0:
            raise OptimizeStop("No data to fit. All pixels are masked")

        # variable to store covariance matrix if it is ever computed
        self._covariance_matrix = None

        # Degrees of freedom
        if ndf is None:
            if self.mask is None:
                self.ndf = max(1.0, len(self.Y) - len(self.current_state))
            else:
                self.ndf = max(1.0, torch.sum(self.mask).item() - len(self.current_state))
        else:
            self.ndf = ndf

    def Lup(self):
        """
        Increases the damping parameter for more gradient-like steps. Used internally.
        """
        self.L = min(1e9, self.L * self._Lup)

    def Ldn(self):
        """
        Decreases the damping parameter for more Gauss-Newton like steps. Used internally.
        """
        self.L = max(1e-9, self.L / self._Ldn)

    @torch.no_grad()
    def step(self, chi2) -> torch.Tensor:
        """Performs one step of the LM algorithm. Computes Jacobian, infers
        hessian and gradient, solves for step vector and iterates on
        damping parameter magnitude until a step with some improvement
        in chi2 is found. Used internally.

        """
        Y0 = self.forward(parameters=self.current_state).flatten("data")
        J = self.jacobian(parameters=self.current_state).flatten("data")
        r = self._r(Y0, self.Y, self.W)
        self.hess = self._hess(J, self.W)
        self.grad = self._grad(J, self.W, Y0, self.Y)
        init_chi2 = chi2
        nostep = True
        best = (torch.zeros_like(self.current_state), init_chi2, self.L)
        scarry_best = (None, init_chi2, self.L)
        direction = "none"
        iteration = 0
        d = 0.1
        for iteration in range(self.max_step_iter):
            # In a scenario where LM is having a hard time proposing a good step, but the damping is really low, just jump up to normal damping levels
            if iteration > self.max_step_iter / 2 and self.L < 1e-3:
                self.L = 1.0

            # compute LM update step
            h = self._h(self.L, self.grad, self.hess)

            # Compute goedesic acceleration
            Y1 = self.forward(parameters=self.current_state + d * h).flatten("data")

            rh = self._r(Y1, self.Y, self.W)

            rpp = self._rpp(J, d, rh - r, self.W, h)

            if self.L > 1e-4:
                a = -self._h(self.L, rpp, self.hess) / 2
            else:
                a = torch.zeros_like(h)

            # Evaluate new step
            ha = h + a * self.acceleration
            Y1 = self.forward(parameters=self.current_state + ha).flatten("data")

            # Compute and report chi^2
            chi2 = self._chi2(Y1.detach()).item()
            if self.verbose > 1:
                AP_config.ap_logger.info(f"sub step L: {self.L}, Chi^2/DoF: {chi2}")

            # Skip if chi^2 is nan
            if not np.isfinite(chi2):
                if self.verbose > 1:
                    AP_config.ap_logger.info("Skip due to non-finite values")
                self.Lup()
                if direction == "better":
                    break
                direction = "worse"
                continue

            # Keep track of chi^2 improvement even if it fails curvature test
            if chi2 <= scarry_best[1]:
                scarry_best = (ha, chi2, self.L)
                nostep = False

            # Check for high curvature, in which case linear approximation is not valid. avoid this step
            rho = torch.linalg.norm(a) / torch.linalg.norm(h)
            if rho > self.curvature_limit:
                if self.verbose > 1:
                    AP_config.ap_logger.info("Skip due to large curvature")
                self.Lup()
                if direction == "better":
                    break
                direction = "worse"
                continue

            # Check for Chi^2 improvement
            if chi2 < best[1]:
                if self.verbose > 1:
                    AP_config.ap_logger.info("new best chi^2")
                best = (ha, chi2, self.L)
                nostep = False
                self.Ldn()
                if self.L <= 1e-8 or direction == "worse":
                    break
                direction = "better"
            elif chi2 > best[1] and direction in ["none", "worse"]:
                if self.verbose > 1:
                    AP_config.ap_logger.info("chi^2 is worse")
                self.Lup()
                if self.L == 1e9:
                    break
                direction = "worse"
            else:
                break

            # If a step substantially improves the chi^2, stop searching for better step, simply exit the loop and accept the good step
            if (best[1] - init_chi2) / init_chi2 < -0.1:
                if self.verbose > 1:
                    AP_config.ap_logger.info("Large step taken, ending search for good step")
                break

        if nostep:
            if scarry_best[0] is not None:
                if self.verbose > 1:
                    AP_config.ap_logger.warning(
                        "no low curvature step found, taking high curvature step"
                    )
                return scarry_best
            raise OptimizeStop("Could not find step to improve chi^2")

        return best

    @staticmethod
    @torch.no_grad()
    def _h(L, grad, hess) -> torch.Tensor:
        I = torch.eye(len(grad), dtype=grad.dtype, device=grad.device)
        D = torch.ones_like(hess) - I
        # Alternate damping scheme
        # (hess + 1e-2 * L**2 * I) * (1 + L**2 * I) ** 2 / (1 + L**2),
        h = torch.linalg.solve(
            hess * (I + D / (1 + L)) + L * I * (1 + torch.diag(hess)),
            grad,
        )

        return h

    @torch.no_grad()
    def _chi2(self, Ypred) -> torch.Tensor:
        if self.mask is None:
            return torch.sum(self.W * (self.Y - Ypred) ** 2) / self.ndf
        else:
            return torch.sum((self.W * (self.Y - Ypred) ** 2)[self.mask]) / self.ndf

    @torch.no_grad()
    def _r(self, Y, Ypred, W) -> torch.Tensor:
        if self.mask is None:
            return W * (Y - Ypred)
        else:
            return W[self.mask] * (Y[self.mask] - Ypred[self.mask])

    @torch.no_grad()
    def _hess(self, J, W) -> torch.Tensor:
        if self.mask is None:
            return J.T @ (W.view(len(W), -1) * J)
        else:
            return J[self.mask].T @ (W[self.mask].view(len(W[self.mask]), -1) * J[self.mask])

    @torch.no_grad()
    def _grad(self, J, W, Y, Ypred) -> torch.Tensor:
        if self.mask is None:
            return -J.T @ self._r(Y, Ypred, W)
        else:
            return -J[self.mask].T @ self._r(Y, Ypred, W)

    @torch.no_grad()
    def _rpp(self, J, d, dr, W, h):
        if self.mask is None:
            return J.T @ ((2 / d) * ((dr / d - W * (J @ h))))
        else:
            return J[self.mask].T @ ((2 / d) * ((dr / d - W[self.mask] * (J[self.mask] @ h))))

    @torch.no_grad()
    def update_hess_grad(self, natural=False) -> None:
        """Updates the stored hessian matrix and gradient vector. This can be
        used to compute the quantities in their natural parameter
        representation. During normal optimization the hessian and
        gradient are computed in a re-mapped parameter space where
        parameters are defined form -inf to inf.

        """
        if natural:
            J = self.jacobian_natural(
                parameters=self.model.parameters.vector_transform_rep_to_val(self.current_state)
            ).flatten("data")
        else:
            J = self.jacobian(parameters=self.current_state).flatten("data")
        Ypred = self.forward(parameters=self.current_state).flatten("data")
        self.hess = self._hess(J, self.W)
        self.grad = self._grad(J, self.W, self.Y, Ypred)

    @torch.no_grad()
    def fit(self) -> BaseOptimizer:
        """This performs the fitting operation. It iterates the LM step
        function until convergence is reached. Includes a message
        after fitting to indicate how the fitting exited. Typically if
        the message returns a "success" then the algorithm found a
        minimum. This may be the desired solution, or a pathological
        local minimum, this often depends on the initial conditions.

        """

        if len(self.current_state) == 0:
            if self.verbose > 0:
                AP_config.ap_logger.warning("No parameters to optimize. Exiting fit")
            return self

        self._covariance_matrix = None
        self.loss_history = [
            self._chi2(self.forward(parameters=self.current_state).flatten("data")).item()
        ]
        self.L_history = [self.L]
        self.lambda_history = [self.current_state.detach().clone().cpu().numpy()]

        for iteration in range(self.max_iter):
            if self.verbose > 0:
                AP_config.ap_logger.info(f"Chi^2/DoF: {self.loss_history[-1]}, L: {self.L}")
            try:
                res = self.step(chi2=self.loss_history[-1])
            except OptimizeStop:
                if self.verbose > 0:
                    AP_config.ap_logger.warning("Could not find step to improve Chi^2, stopping")
                self.message = self.message + "fail. Could not find step to improve Chi^2"
                break

            self.L = res[2]
            self.current_state = (self.current_state + res[0]).detach()
            self.L_history.append(self.L)
            self.loss_history.append(res[1])
            self.lambda_history.append(self.current_state.detach().clone().cpu().numpy())

            self.Ldn()

            if len(self.loss_history) >= 3:
                if (self.loss_history[-3] - self.loss_history[-1]) / self.loss_history[
                    -1
                ] < self.relative_tolerance and self.L < 0.1:
                    self.message = self.message + "success"
                    break
            if len(self.loss_history) > 10:
                if (self.loss_history[-10] - self.loss_history[-1]) / self.loss_history[
                    -1
                ] < self.relative_tolerance:
                    self.message = (
                        self.message + "success by immobility. Convergence not guaranteed"
                    )
                    break

        else:
            self.message = self.message + "fail. Maximum iterations"

        if self.verbose > 0:
            AP_config.ap_logger.info(
                f"Final Chi^2/DoF: {self.loss_history[-1]}, L: {self.L_history[-1]}. Converged: {self.message}"
            )
        self.model.parameters.vector_set_representation(self.res())

        return self

    @property
    @torch.no_grad()
    def covariance_matrix(self) -> torch.Tensor:
        """The covariance matrix for the model at the current
        parameters. This can be used to construct a full Gaussian PDF
        for the parameters using: :math:`\\mathcal{N}(\\mu,\\Sigma)`
        where :math:`\\mu` is the optimized parameters and
        :math:`\\Sigma` is the covariance matrix.

        """

        if self._covariance_matrix is not None:
            return self._covariance_matrix
        self.update_hess_grad(natural=True)
        try:
            self._covariance_matrix = torch.linalg.inv(self.hess)
        except:
            AP_config.ap_logger.warning(
                "WARNING: Hessian is singular, likely at least one model is non-physical. Will massage Hessian to continue but results should be inspected."
            )
            self.hess += torch.eye(
                len(self.grad), dtype=AP_config.ap_dtype, device=AP_config.ap_device
            ) * (torch.diag(self.hess) == 0)
            self._covariance_matrix = torch.linalg.inv(self.hess)
        return self._covariance_matrix

    @torch.no_grad()
    def update_uncertainty(self) -> None:
        """Call this function after optimization to set the uncertainties for
        the parameters. This will use the diagonal of the covariance
        matrix to update the uncertainties. See the covariance_matrix
        function for the full representation of the uncertainties.

        """
        # set the uncertainty for each parameter
        cov = self.covariance_matrix
        if torch.all(torch.isfinite(cov)):
            try:
                self.model.parameters.vector_set_uncertainty(torch.sqrt(torch.abs(torch.diag(cov))))
            except RuntimeError as e:
                AP_config.ap_logger.warning(f"Unable to update uncertainty due to: {e}")
        else:
            AP_config.ap_logger.warning(
                "Unable to update uncertainty due to non finite covariance matrix"
            )



from .base import *
from .lm import *
from .oldlm import *
from .gradient import *
from .iterative import *
from .minifit import *

try:
    from .hmc import *
    from .nuts import *
except AssertionError as e:
    print("Could not load HMC or NUTS due to:", str(e))
from .mhmcmc import *

"""
base: This module defines the base class BaseOptimizer,
      which is used as the parent class for all optimization algorithms in AstroPhot.
      This module contains helper functions used across multiple optimization algorithms,
      such as computing gradients and making copies of models.

LM: This module defines the class LM,
    which uses the Levenberg-Marquardt algorithm to perform optimization.
    This algorithm adjusts the learning rate at each step to find the optimal value.

Grad: This module defines the class Gradient-Optimizer,
      which uses a simple gradient descent algorithm to perform optimization.
      This algorithm adjusts the learning rate at each step to find the optimal value.

Iterative: This module defines the class Iter,
            which uses an iterative algorithm to perform Optimization.
            This algorithm repeatedly fits each model individually until they all converge.

"""



import numpy as np
from scipy.optimize import minimize

from ..interpolate import point_Lanczos
from ... import AP_config


def center_of_mass(center, image, window=None):
    """Iterative light weighted center of mass optimization. Each step
    determines the light weighted center of mass within a small
    window. The new center is used to create a new window. This
    continues until the center no longer updates or an image boundary
    is reached.

    """
    if window is None:
        window = max(min(int(min(image.shape) / 10), 30), 6)
    init_center = center
    window += window % 2
    xx, yy = np.meshgrid(np.arange(window), np.arange(window))
    for iteration in range(100):
        # Determine the image window to calculate COM
        ranges = [
            [int(round(center[0]) - window / 2), int(round(center[0]) + window / 2)],
            [int(round(center[1]) - window / 2), int(round(center[1]) + window / 2)],
        ]
        # Avoid edge of image
        if (
            ranges[0][0] < 0
            or ranges[1][0] < 0
            or ranges[0][1] >= image.shape[0]
            or ranges[1][1] >= image.shape[1]
        ):
            AP_config.ap_logger.warning("Image edge!")
            return init_center

        # Compute COM
        denom = np.sum(image[ranges[0][0] : ranges[0][1], ranges[1][0] : ranges[1][1]])
        new_center = [
            ranges[0][0]
            + np.sum(image[ranges[0][0] : ranges[0][1], ranges[1][0] : ranges[1][1]] * yy) / denom,
            ranges[1][0]
            + np.sum(image[ranges[0][0] : ranges[0][1], ranges[1][0] : ranges[1][1]] * xx) / denom,
        ]
        new_center = np.array(new_center)
        # Check for convergence
        if np.sum(np.abs(np.array(center) - new_center)) < 0.1:
            break

        center = new_center

    return center


def GaussianDensity_Peak(center, image, window=10, std=0.5):
    init_center = center
    window += window % 2

    def _add_flux(c):
        r = np.round(center)
        xx, yy = np.meshgrid(
            np.arange(r[0] - window / 2, r[0] + window / 2 + 1) - c[0],
            np.arange(r[1] - window / 2, r[1] + window / 2 + 1) - c[1],
        )
        rr2 = xx**2 + yy**2
        f = image[
            int(r[1] - window / 2) : int(r[1] + window / 2 + 1),
            int(r[0] - window / 2) : int(r[0] + window / 2 + 1),
        ]
        return -np.sum(np.exp(-rr2 / (2 * std)) * f)

    res = minimize(_add_flux, x0=center)
    return res.x


def Lanczos_peak(center, image, Lanczos_scale=3):
    best = [np.inf, None]
    for dx in np.arange(-3, 4):
        for dy in np.arange(-3, 4):
            res = minimize(
                lambda x: -point_Lanczos(image, x[0], x[1], scale=Lanczos_scale),
                x0=(center[0] + dx, center[1] + dy),
                method="Nelder-Mead",
            )
            if res.fun < best[0]:
                best[0] = res.fun
                best[1] = res.x
    return best[1]



from typing import List, Optional, Union

import torch
import numpy as np

from .image_object import Image
from .image_header import Image_Header
from .model_image import Model_Image
from .jacobian_image import Jacobian_Image
from astropy.io import fits
from .. import AP_config
from ..errors import SpecificationConflict

__all__ = ["PSF_Image"]


class PSF_Image(Image):
    """Image object which represents a model of PSF (Point Spread Function).

    PSF_Image inherits from the base Image class and represents the model of a point spread function.
    The point spread function characterizes the response of an imaging system to a point source or point object.

    The shape of the PSF data must be odd.

    Attributes:
        data (torch.Tensor): The image data of the PSF.
        identity (str): The identity of the image. Default is None.

    Methods:
        psf_border_int: Calculates and returns the convolution border size of the PSF image in integer format.
        psf_border: Calculates and returns the convolution border size of the PSF image in the units of pixelscale.
        _save_image_list: Saves the image list to the PSF HDU header.
        reduce: Reduces the size of the image using a given scale factor.
    """

    has_mask = False
    has_variance = False

    def __init__(self, *args, **kwargs):
        """
        Initializes the PSF_Image class.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
                band (str, optional): The band of the image. Default is None.
        """
        super().__init__(*args, **kwargs)

        self.window.reference_radec = (0, 0)
        self.window.reference_planexy = (0, 0)
        self.window.reference_imageij = np.flip(np.array(self.data.shape, dtype=float) - 1.0) / 2
        self.window.reference_imagexy = (0, 0)

    def set_data(self, data: Union[torch.Tensor, np.ndarray], require_shape: bool = True):
        super().set_data(data=data, require_shape=require_shape)

        if torch.any((torch.tensor(self.data.shape) % 2) != 1):
            raise SpecificationConflict(f"psf must have odd shape, not {self.data.shape}")
        if torch.any(self.data < 0):
            AP_config.ap_logger.warning("psf data should be non-negative")

    def normalize(self):
        """Normalizes the PSF image to have a sum of 1."""
        self.data /= torch.sum(self.data)

    @property
    def mask(self):
        return torch.zeros_like(self.data, dtype=bool)

    @property
    def psf_border_int(self):
        """Calculates and returns the border size of the PSF image in integer
        format. This is the border used for padding before convolution.

        Returns:
            torch.Tensor: The border size of the PSF image in integer format.

        """
        return torch.ceil(
            (
                1
                + torch.flip(
                    torch.tensor(
                        self.data.shape,
                        dtype=AP_config.ap_dtype,
                        device=AP_config.ap_device,
                    ),
                    (0,),
                )
            )
            / 2
        ).int()

    def _save_image_list(self, image_list):
        """Saves the image list to the PSF HDU header.

        Args:
            image_list (list): The list of images to be saved.
            psf_header (astropy.io.fits.Header): The header of the PSF HDU.
        """
        img_header = self.header._save_image_list()
        img_header["IMAGE"] = "PSF"
        image_list.append(fits.ImageHDU(self.data.detach().cpu().numpy(), header=img_header))

    def jacobian_image(
        self,
        parameters: Optional[List[str]] = None,
        data: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        """
        Construct a blank `Jacobian_Image` object formatted like this current `PSF_Image` object. Mostly used internally.
        """
        if parameters is None:
            data = None
            parameters = []
        elif data is None:
            data = torch.zeros(
                (*self.data.shape, len(parameters)),
                dtype=AP_config.ap_dtype,
                device=AP_config.ap_device,
            )
        return Jacobian_Image(
            parameters=parameters,
            target_identity=self.identity,
            data=data,
            header=self.header,
            **kwargs,
        )

    def model_image(self, data: Optional[torch.Tensor] = None, **kwargs):
        """
        Construct a blank `Model_Image` object formatted like this current `Target_Image` object. Mostly used internally.
        """
        return Model_Image(
            data=torch.zeros_like(self.data) if data is None else data,
            header=self.header,
            target_identity=self.identity,
            **kwargs,
        )

    def expand(self, padding):
        raise NotImplementedError("expand not available for PSF_Image")

    def get_fits_state(self):
        states = [{}]
        states[0]["DATA"] = self.data.detach().cpu().numpy()
        states[0]["HEADER"] = self.header.get_fits_state()
        states[0]["HEADER"]["IMAGE"] = "PSF"
        return states

    def set_fits_state(self, states):
        for state in states:
            if state["HEADER"]["IMAGE"] == "PSF":
                self.set_data(np.array(state["DATA"], dtype=np.float64), require_shape=False)
                self.header = Image_Header(fits_state=state["HEADER"])
                break



from typing import Optional, Union, Any

import torch
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS as AstropyWCS

from .window_object import Window
from .. import AP_config

__all__ = ["Image_Header"]


class Image_Header:
    """Store meta-information for images to be used in AstroPhot.

    The Image_Header object stores all meta information which tells
    AstroPhot what is contained in an image array of pixels. This
    includes coordinate systems and how to transform between them (see
    :doc:`coordinates`). The image header will also know the image
    zeropoint if that data is available.

    Args:
      window : Window or None, optional
          A Window object defining the area of the image in the coordinate
          systems. Default is None.
      filename : str or None, optional
          The name of a file containing the image data. Default is None.
      zeropoint : float or None, optional
          The image's zeropoint, used for flux calibration. Default is None.
      metadata : dict or None, optional
          Any information the user wishes to associate with this image, stored in a python dictionary. Default is None.

    """

    north = np.pi / 2.0

    def __init__(
        self,
        *,
        data_shape: Optional[torch.Tensor] = None,
        wcs: Optional[AstropyWCS] = None,
        window: Optional[Window] = None,
        filename: Optional[str] = None,
        zeropoint: Optional[Union[float, torch.Tensor]] = None,
        metadata: Optional[dict] = None,
        identity: str = None,
        state: Optional[dict] = None,
        fits_state: Optional[dict] = None,
        **kwargs: Any,
    ) -> None:
        # Record identity
        if identity is None:
            self.identity = str(id(self))
        else:
            self.identity = identity

        # set Zeropoint
        self.zeropoint = zeropoint

        # set metadata for the image
        self.metadata = metadata

        if filename is not None:
            self.load(filename)
            return
        elif state is not None:
            self.set_state(state)
            return
        elif fits_state is not None:
            self.set_fits_state(fits_state)
            return

        # Set Window
        if window is None:
            data_shape = torch.as_tensor(data_shape, dtype=torch.int32, device=AP_config.ap_device)
            # If window is not provided, create one based on provided information
            self.window = Window(
                pixel_shape=torch.flip(data_shape, (0,)),
                wcs=wcs,
                **kwargs,
            )
        else:
            # When the Window object is provided
            self.window = window

    @property
    def zeropoint(self):
        """The photometric zeropoint of the image, used as a flux reference
        point.

        """
        return self._zeropoint

    @zeropoint.setter
    def zeropoint(self, zp):
        if zp is None:
            self._zeropoint = None
            return

        self._zeropoint = (
            torch.as_tensor(zp, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
            .clone()
            .detach()
        )

    @property
    def origin(self) -> torch.Tensor:
        """
        Returns the location of the origin (pixel coordinate -0.5, -0.5) of the image window in the tangent plane (arcsec).

        Returns:
            torch.Tensor: A 1D tensor of shape (2,) containing the (x, y) coordinates of the origin.
        """
        return self.window.origin

    @property
    def shape(self) -> torch.Tensor:
        """
        Returns the shape (size) of the image window (arcsec, arcsec).

        Returns:
                torch.Tensor: A 1D tensor of shape (2,) containing the (width, height) of the window in arcsec.
        """
        return self.window.shape

    @property
    def center(self) -> torch.Tensor:
        """
        Returns the center of the image window (arcsec).

        Returns:
            torch.Tensor: A 1D tensor of shape (2,) containing the (x, y) coordinates of the center.
        """
        return self.window.center

    def world_to_plane(self, *args, **kwargs):
        return self.window.world_to_plane(*args, **kwargs)

    def plane_to_world(self, *args, **kwargs):
        return self.window.plane_to_world(*args, **kwargs)

    def plane_to_pixel(self, *args, **kwargs):
        return self.window.plane_to_pixel(*args, **kwargs)

    def pixel_to_plane(self, *args, **kwargs):
        return self.window.pixel_to_plane(*args, **kwargs)

    def plane_to_pixel_delta(self, *args, **kwargs):
        return self.window.plane_to_pixel_delta(*args, **kwargs)

    def pixel_to_plane_delta(self, *args, **kwargs):
        return self.window.pixel_to_plane_delta(*args, **kwargs)

    def world_to_pixel(self, *args, **kwargs):
        return self.window.world_to_pixel(*args, **kwargs)

    def pixel_to_world(self, *args, **kwargs):
        return self.window.pixel_to_world(*args, **kwargs)

    def get_coordinate_meshgrid(self):
        return self.window.get_coordinate_meshgrid()

    def get_coordinate_corner_meshgrid(self):
        return self.window.get_coordinate_corner_meshgrid()

    def get_coordinate_simps_meshgrid(self):
        return self.window.get_coordinate_simps_meshgrid()

    @property
    def pixelscale(self):
        return self.window.pixelscale

    @property
    def pixel_length(self):
        return self.window.pixel_length

    @property
    def pixel_area(self):
        return self.window.pixel_area

    def shift(self, shift):
        """Adjust the position of the image described by the header. This will
        not adjust the data represented by the header, only the
        coordinate system that maps pixel coordinates to the plane
        coordinates.

        """
        self.window.shift(shift)

    def pixel_shift(self, shift):
        self.window.pixel_shift(shift)

    def copy(self, **kwargs):
        """Produce a copy of this image with all of the same properties. This
        can be used when one wishes to make temporary modifications to
        an image and then will want the original again.

        """
        copy_kwargs = {
            "zeropoint": self.zeropoint,
            "metadata": self.metadata,
            "window": self.window.copy(),
            "identity": self.identity,
        }
        copy_kwargs.update(kwargs)
        return self.__class__(**copy_kwargs)

    def get_window(self, window, **kwargs):
        """Get a sub-region of the image as defined by a window on the sky."""
        copy_kwargs = {
            "window": self.window & window,
        }
        copy_kwargs.update(kwargs)
        return self.copy(**copy_kwargs)

    def to(self, dtype=None, device=None):
        if dtype is None:
            dtype = AP_config.ap_dtype
        if device is None:
            device = AP_config.ap_device
        self.window.to(dtype=dtype, device=device)
        if self.zeropoint is not None:
            self.zeropoint.to(dtype=dtype, device=device)
        return self

    def crop(self, pixels):  # fixme data_shape?
        """Reduce the size of an image by cropping some number of pixels off
        the borders. If pixels is a single value, that many pixels are
        cropped off all sides. If pixels is two values then a different
        crop is done in x vs y. If pixels is four values then crop on
        all sides are specified explicitly.

        formatted as:
        [crop all sides] or
        [crop x, crop y] or
        [crop x low, crop y low, crop x high, crop y high]

        """
        self.window.crop_pixel(pixels)
        return self

    def rescale_pixel(self, scale: int, **kwargs):
        if scale == 1:
            return self

        return self.copy(
            window=self.window.rescale_pixel(scale),
            **kwargs,
        )

    def get_state(self):
        """Returns a dictionary with necessary information to recreate the
        Image_Header object.

        """
        state = {}
        if self.zeropoint is not None:
            state["zeropoint"] = self.zeropoint.item()
        state["window"] = self.window.get_state()
        if self.metadata is not None:
            state["metadata"] = self.metadata
        return state

    def set_state(self, state):
        self.zeropoint = state.get("zeropoint", self.zeropoint)
        self.window = Window(state=state["window"])
        self.metadata = state.get("metadata", self.metadata)

    def get_fits_state(self):
        state = {}
        state.update(self.window.get_fits_state())
        if self.zeropoint is not None:
            state["ZEROPNT"] = str(self.zeropoint.detach().cpu().item())
        if self.metadata is not None:
            state["METADATA"] = str(self.metadata)
        return state

    def set_fits_state(self, state):
        """
        Updates the state of the Image_Header using information saved in a FITS header (more generally, a properly formatted dictionary will also work but not yet).
        """
        self.zeropoint = eval(state.get("ZEROPNT", "None"))
        self.metadata = state.get("METADATA", None)
        self.window = Window(fits_state=state)

    def _save_image_list(self):
        """
        Constructs a FITS header object which has the necessary information to recreate the Image_Header object.
        """
        img_header = fits.Header()
        img_header["IMAGE"] = "PRIMARY"
        img_header["WINDOW"] = str(self.window.get_state())
        if self.zeropoint is not None:
            img_header["ZEROPNT"] = str(self.zeropoint.detach().cpu().item())
        if self.metadata is not None:
            img_header["METADATA"] = str(self.metadata)
        return img_header

    def save(self, filename=None, overwrite=True):
        """
        Save header to a FITS file.
        """
        image_list = self._save_image_list()
        hdul = fits.HDUList(image_list)
        if filename is not None:
            hdul.writeto(filename, overwrite=overwrite)
        return hdul

    def load(self, filename):
        """
        load header from a FITS file.
        """
        hdul = fits.open(filename)
        for hdu in hdul:
            if "IMAGE" in hdu.header and hdu.header["IMAGE"] == "PRIMARY":
                self.set_fits_state(hdu.header)
                break
        return hdul

    def __str__(self):
        state = self.get_state()
        state.update(self.window.get_state())
        keys = ["pixel_shape", "pixelscale", "reference_imageij", "reference_imagexy"]
        if "zeropoint" in state:
            keys.append("zeropoint")
        if "metadata" in state:
            keys.append("metadata")
        return "\n".join(f"{key}: {state[key]}" for key in keys)

    def __repr__(self):
        state = self.get_state()
        state.update(self.window.get_state())
        return "\n".join(f"{key}: {state[key]}" for key in sorted(state.keys()))



import numpy as np
from scipy.stats import iqr
import torch

from .sky_model_object import Sky_Model
from ._shared_methods import select_target
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..param import Param_Unlock, Param_SoftLimits

__all__ = ["Plane_Sky"]


class Plane_Sky(Sky_Model):
    """Sky background model using a tilted plane for the sky flux. The brightness for each pixel is defined as:

    I(X, Y) = S + X*dx + Y*dy

    where I(X,Y) is the brightness as a function of image position X Y,
    S is the central sky brightness value, and dx dy are the slopes of
    the sky brightness plane.

    Parameters:
        sky: central sky brightness value
        delta: Tensor for slope of the sky brightness in each image dimension

    """

    model_type = f"plane {Sky_Model.model_type}"
    parameter_specs = {
        "F": {"units": "flux/arcsec^2"},
        "delta": {"units": "flux/arcsec"},
    }
    _parameter_order = Sky_Model._parameter_order + ("F", "delta")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        with Param_Unlock(parameters["F"]), Param_SoftLimits(parameters["F"]):
            if parameters["F"].value is None:
                parameters["F"].value = (
                    np.median(target[self.window].data.detach().cpu().numpy())
                    / target.pixel_area.item()
                )
            if parameters["F"].uncertainty is None:
                parameters["F"].uncertainty = (
                    iqr(
                        target[self.window].data.detach().cpu().numpy(),
                        rng=(31.731 / 2, 100 - 31.731 / 2),
                    )
                    / (2.0)
                ) / np.sqrt(np.prod(self.window.shape.detach().cpu().numpy()))
        with Param_Unlock(parameters["delta"]), Param_SoftLimits(parameters["delta"]):
            if parameters["delta"].value is None:
                parameters["delta"].value = [0.0, 0.0]
                parameters["delta"].uncertainty = [
                    self.default_uncertainty,
                    self.default_uncertainty,
                ]

    @default_internal
    def evaluate_model(self, X=None, Y=None, image=None, parameters=None, **kwargs):
        if X is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
        return (
            image.pixel_area * parameters["F"].value
            + X * parameters["delta"].value[0]
            + Y * parameters["delta"].value[1]
        )



from .segmentation_map import *
from .initialize import isophotes
from .center import center_of_mass, GaussianDensity_Peak, Lanczos_peak
from .construct_psf import gaussian_psf, moffat_psf, construct_psf

__all__ = (
    "isophotes",
    "center_of_mass",
    "GaussianDensity_Peak",
    "Lanczos_peak",
    "gaussian_psf",
    "moffat_psf",
    "construct_psf",
    "centroids_from_segmentation_map",
    "PA_from_segmentation_map",
    "q_from_segmentation_map",
    "windows_from_segmentation_map",
    "scale_windows",
    "filter_windows",
    "transfer_windows",
)



import torch
import numpy as np
from .conversions.functions import sersic_n_to_b
from .interpolate import cubic_spline_torch


def sersic_torch(R, n, Re, Ie):
    """Seric 1d profile function, specifically designed for pytorch
    operations

    Parameters:
        R: Radii tensor at which to evaluate the sersic function
        n: sersic index restricted to n > 0.36
        Re: Effective radius in the same units as R
        Ie: Effective surface density
    """
    bn = sersic_n_to_b(n)
    return Ie * torch.exp(-bn * (torch.pow(R / Re, 1 / n) - 1))


def sersic_np(R, n, Re, Ie):
    """Sersic 1d profile function, works more generally with numpy
    operations. In the event that impossible values are passed to the
    function it returns large values to guide optimizers away from
    such values.

    Parameters:
        R: Radii array at which to evaluate the sersic function
        n: sersic index restricted to n > 0.36
        Re: Effective radius in the same units as R
        Ie: Effective surface density
    """
    if np.any(np.array([n, Re, Ie]) <= 0):
        return np.ones(len(R)) * 1e6
    bn = sersic_n_to_b(n)
    return Ie * np.exp(-bn * ((R / Re) ** (1 / n) - 1))


def gaussian_torch(R, sigma, I0):
    """Gaussian 1d profile function, specifically designed for pytorch
    operations.

    Parameters:
        R: Radii tensor at which to evaluate the sersic function
        sigma: standard deviation of the gaussian in the same units as R
        I0: central surface density
    """
    return (I0 / torch.sqrt(2 * np.pi * sigma**2)) * torch.exp(-0.5 * torch.pow(R / sigma, 2))


def gaussian_np(R, sigma, I0):
    """Gaussian 1d profile function, works more generally with numpy
    operations.

    Parameters:
        R: Radii array at which to evaluate the sersic function
        sigma: standard deviation of the gaussian in the same units as R
        I0: central surface density
    """
    return (I0 / np.sqrt(2 * np.pi * sigma**2)) * np.exp(-0.5 * ((R / sigma) ** 2))


def exponential_torch(R, Re, Ie):
    """Exponential 1d profile function, specifically designed for pytorch
    operations.

    Parameters:
        R: Radii tensor at which to evaluate the sersic function
        Re: Effective radius in the same units as R
        Ie: Effective surface density
    """
    return Ie * torch.exp(
        -sersic_n_to_b(torch.tensor(1.0, dtype=R.dtype, device=R.device)) * ((R / Re) - 1.0)
    )


def exponential_np(R, Ie, Re):
    """Exponential 1d profile function, works more generally with numpy
    operations.

    Parameters:
        R: Radii array at which to evaluate the sersic function
        Re: Effective radius in the same units as R
        Ie: Effective surface density
    """
    return Ie * np.exp(-sersic_n_to_b(1.0) * (R / Re - 1.0))


def moffat_torch(R, n, Rd, I0):
    """Moffat 1d profile function, specifically designed for pytorch
    operations

    Parameters:
        R: Radii tensor at which to evaluate the moffat function
        n: concentration index
        Rd: scale length in the same units as R
        I0: central surface density

    """
    return I0 / (1 + (R / Rd) ** 2) ** n


def moffat_np(R, n, Rd, I0):
    """Moffat 1d profile function, works with numpy operations.

    Parameters:
        R: Radii tensor at which to evaluate the moffat function
        n: concentration index
        Rd: scale length in the same units as R
        I0: central surface density

    """
    return I0 / (1 + (R / Rd) ** 2) ** n


def nuker_torch(R, Rb, Ib, alpha, beta, gamma):
    """Nuker 1d profile function, specifically designed for pytorch
    operations

    Parameters:
        R: Radii tensor at which to evaluate the nuker function
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """
    return (
        Ib
        * (2 ** ((beta - gamma) / alpha))
        * ((R / Rb) ** (-gamma))
        * ((1 + (R / Rb) ** alpha) ** ((gamma - beta) / alpha))
    )


def nuker_np(R, Rb, Ib, alpha, beta, gamma):
    """Nuker 1d profile function, works with numpy functions

    Parameters:
        R: Radii tensor at which to evaluate the nuker function
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """
    return (
        Ib
        * (2 ** ((beta - gamma) / alpha))
        * ((R / Rb) ** (-gamma))
        * ((1 + (R / Rb) ** alpha) ** ((gamma - beta) / alpha))
    )


def spline_torch(R, profR, profI, extend):
    """Spline 1d profile function, cubic spline between points up
    to second last point beyond which is linear, specifically designed
    for pytorch.

    Parameters:
        R: Radii tensor at which to evaluate the sersic function
        profR: radius values for the surface density profile in the same units as R
        profI: surface density values for the surface density profile
    """
    I = cubic_spline_torch(profR, profI, R.view(-1), extend="none").view(*R.shape)
    res = torch.zeros_like(I)
    res[R <= profR[-1]] = 10 ** (I[R <= profR[-1]])
    if extend:
        res[R > profR[-1]] = 10 ** (
            profI[-2]
            + (R[R > profR[-1]] - profR[-2]) * ((profI[-1] - profI[-2]) / (profR[-1] - profR[-2]))
        )
    else:
        res[R > profR[-1]] = 0
    return res



import unittest

import torch
import numpy as np

import astrophot as ap
from utils import make_basic_sersic

######################################################################
# Fit Objects
######################################################################


class TestComponentModelFits(unittest.TestCase):
    def test_sersic_fit_grad(self):
        """
        Simply test that the gradient optimizer changes the parameters
        """
        np.random.seed(12345)
        N = 50
        Width = 20
        shape = (N + 10, N)
        true_params = [2, 5, 10, -3, 5, 0.7, np.pi / 4]
        IXX, IYY = np.meshgrid(
            np.linspace(-Width, Width, shape[1]), np.linspace(-Width, Width, shape[0])
        )
        QPAXX, QPAYY = ap.utils.conversions.coordinates.Axis_Ratio_Cartesian_np(
            true_params[5], IXX - true_params[3], IYY - true_params[4], true_params[6]
        )
        Z0 = ap.utils.parametric_profiles.sersic_np(
            np.sqrt(QPAXX**2 + QPAYY**2),
            true_params[0],
            true_params[1],
            true_params[2],
        ) + np.random.normal(loc=0, scale=0.1, size=shape)
        tar = ap.image.Target_Image(
            data=Z0,
            pixelscale=0.8,
            variance=np.ones(Z0.shape) * (0.1**2),
        )

        mod = ap.models.Sersic_Galaxy(
            name="sersic model",
            target=tar,
            parameters={
                "center": [-3.2 + N / 2, 5.1 + (N + 10) / 2],
                "q": 0.6,
                "PA": np.pi / 4,
                "n": 2,
                "Re": 5,
                "Ie": 10,
            },
        )

        self.assertFalse(mod.locked, "default model should not be locked")

        mod.initialize()

        mod_initparams = {}
        for p in mod.parameters:
            mod_initparams[p.name] = np.copy(p.vector_representation().detach().cpu().numpy())

        res = ap.fit.Grad(model=mod, max_iter=10).fit()

        for p in mod.parameters:
            self.assertFalse(
                np.any(p.vector_representation().detach().cpu().numpy() == mod_initparams[p.name]),
                f"parameter {p.name} should update with optimization",
            )

    def test_sersic_fit_lm(self):
        """
        Test sersic fitting with entirely independent sersic sampling at 10x resolution.
        """
        N = 50
        pixelscale = 0.8
        shape = (N + 10, N)
        true_params = {
            "center": [
                shape[0] * pixelscale / 2 - 3.35,
                shape[1] * pixelscale / 2 + 5.35,
            ],
            "n": 1,
            "Re": 20,
            "Ie": 0.0,
            "q": 0.7,
            "PA": np.pi / 4,
        }
        tar = make_basic_sersic(
            N=shape[0],
            M=shape[1],
            pixelscale=pixelscale,
            x=true_params["center"][0],
            y=true_params["center"][1],
            n=true_params["n"],
            Re=true_params["Re"],
            Ie=true_params["Ie"],
            q=true_params["q"],
            PA=true_params["PA"],
        )
        mod = ap.models.AstroPhot_Model(
            name="sersic model",
            model_type="sersic galaxy model",
            target=tar,
            sampling_mode="simpsons",
        )

        mod.initialize()
        ap.AP_config.set_logging_output(stdout=True, filename="AstroPhot.log")
        res = ap.fit.LM(model=mod, verbose=2).fit()
        res.update_uncertainty()

        self.assertAlmostEqual(
            mod["center"].value[0].item() / true_params["center"][0],
            1,
            2,
            "LM should accurately recover parameters in simple cases",
        )
        self.assertAlmostEqual(
            mod["center"].value[1].item() / true_params["center"][1],
            1,
            2,
            "LM should accurately recover parameters in simple cases",
        )
        self.assertAlmostEqual(
            mod["n"].value.item(),
            true_params["n"],
            1,
            msg="LM should accurately recover parameters in simple cases",
        )
        self.assertAlmostEqual(
            (mod["Re"].value.item()) / true_params["Re"],
            1,
            delta=1,
            msg="LM should accurately recover parameters in simple cases",
        )
        self.assertAlmostEqual(
            mod["Ie"].value.item(),
            true_params["Ie"],
            1,
            "LM should accurately recover parameters in simple cases",
        )
        self.assertAlmostEqual(
            mod["PA"].value.item() / true_params["PA"],
            1,
            delta=0.5,
            msg="LM should accurately recover parameters in simple cases",
        )
        self.assertAlmostEqual(
            mod["q"].value.item(),
            true_params["q"],
            1,
            "LM should accurately recover parameters in simple cases",
        )
        cov = res.covariance_matrix


class TestGroupModelFits(unittest.TestCase):
    def test_groupmodel_fit(self):
        """
        Simply test that fitting a group model changes the parameter values
        """
        np.random.seed(12345)
        N = 50
        Width = 20
        shape = (N + 10, N)
        true_params1 = [2, 4, 10, -3, 5, 0.7, np.pi / 4]
        true_params2 = [1.2, 6, 8, 2, -3, 0.5, -np.pi / 4]
        IXX, IYY = np.meshgrid(
            np.linspace(-Width, Width, shape[1]), np.linspace(-Width, Width, shape[0])
        )
        QPAXX, QPAYY = ap.utils.conversions.coordinates.Axis_Ratio_Cartesian_np(
            true_params1[5],
            IXX - true_params1[3],
            IYY - true_params1[4],
            true_params1[6],
        )
        Z0 = ap.utils.parametric_profiles.sersic_np(
            np.sqrt(QPAXX**2 + QPAYY**2),
            true_params1[0],
            true_params1[1],
            true_params1[2],
        )
        QPAXX, QPAYY = ap.utils.conversions.coordinates.Axis_Ratio_Cartesian_np(
            true_params2[5],
            IXX - true_params2[3],
            IYY - true_params2[4],
            true_params2[6],
        )
        Z0 += ap.utils.parametric_profiles.sersic_np(
            np.sqrt(QPAXX**2 + QPAYY**2),
            true_params2[0],
            true_params2[1],
            true_params2[2],
        )
        Z0 += np.random.normal(loc=0, scale=0.1, size=shape)
        tar = ap.image.Target_Image(
            data=Z0,
            pixelscale=0.8,
            variance=np.ones(Z0.shape) * (0.1**2),
        )

        mod1 = ap.models.Sersic_Galaxy(
            name="sersic model 1",
            target=tar,
            parameters={"center": {"value": [-3.2 + N / 2, 5.1 + (N + 10) / 2]}},
        )
        mod2 = ap.models.Sersic_Galaxy(
            name="sersic model 2",
            target=tar,
            parameters={"center": {"value": [2.1 + N / 2, -3.1 + (N + 10) / 2]}},
        )

        smod = ap.models.Group_Model(name="group model", models=[mod1, mod2], target=tar)

        self.assertFalse(smod.locked, "default model should not be locked")

        smod.initialize()

        mod1_initparams = {}
        for p in mod1.parameters:
            mod1_initparams[p.name] = np.copy(p.vector_representation().detach().cpu().numpy())
        mod2_initparams = {}
        for p in mod2.parameters:
            mod2_initparams[p.name] = np.copy(p.vector_representation().detach().cpu().numpy())

        res = ap.fit.Grad(model=smod, max_iter=10).fit()

        for p in mod1.parameters:
            self.assertFalse(
                np.any(p.vector_representation().detach().cpu().numpy() == mod1_initparams[p.name]),
                f"mod1 parameter {p.name} should update with optimization",
            )
        for p in mod2.parameters:
            self.assertFalse(
                np.any(p.vector_representation().detach().cpu().numpy() == mod2_initparams[p.name]),
                f"mod2 parameter {p.name} should update with optimization",
            )


class TestLM(unittest.TestCase):
    def test_lm_creation(self):
        target = make_basic_sersic()
        new_model = ap.models.AstroPhot_Model(
            name="test sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
        )

        LM = ap.fit.LM(new_model, max_iter=10)

        LM.fit()

    def test_chunk_parameter_jacobian(self):
        target = make_basic_sersic()
        new_model = ap.models.AstroPhot_Model(
            name="test sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
            jacobian_chunksize=3,
        )

        LM = ap.fit.LM(new_model, max_iter=10)

        LM.fit()

    def test_chunk_image_jacobian(self):
        target = make_basic_sersic()
        new_model = ap.models.AstroPhot_Model(
            name="test sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
            image_chunksize=15,
        )

        LM = ap.fit.LM(new_model, max_iter=10)

        LM.fit()

    def test_group_fit_step(self):
        np.random.seed(123456)
        tar = make_basic_sersic(N=51, M=51)
        mod1 = ap.models.Sersic_Galaxy(
            name="base model 1",
            target=tar,
            window=[[0, 25], [0, 25]],
            parameters={
                "center": [5, 5],
                "PA": 0,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
        )
        mod2 = ap.models.Sersic_Galaxy(
            name="base model 2",
            target=tar,
            window=[[25, 51], [25, 51]],
            parameters={
                "center": [5, 5],
                "PA": 0,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
        )

        smod = ap.models.AstroPhot_Model(
            name="group model",
            model_type="group model",
            models=[mod1, mod2],
            target=tar,
        )
        vec_init = smod.parameters.vector_values().detach().clone()
        LM = ap.fit.LM(smod, max_iter=1).fit()
        vec_final = smod.parameters.vector_values().detach().clone()
        self.assertFalse(
            torch.all(vec_init == vec_final),
            "LM should update parameters in LM step",
        )


class TestMiniFit(unittest.TestCase):
    def test_minifit(self):
        target = make_basic_sersic()
        new_model = ap.models.AstroPhot_Model(
            name="test sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
        )

        MF = ap.fit.MiniFit(
            new_model, downsample_factor=2, method_quargs={"max_iter": 10}, verbose=1
        )

        MF.fit()


class TestIter(unittest.TestCase):
    def test_iter_basic(self):
        target = make_basic_sersic()
        model_list = []
        model_list.append(
            ap.models.AstroPhot_Model(
                name="basic sersic",
                model_type="sersic galaxy model",
                parameters={
                    "center": [20, 20],
                    "PA": 60 * np.pi / 180,
                    "q": 0.5,
                    "n": 2,
                    "Re": 5,
                    "Ie": 1,
                },
                target=target,
            )
        )
        model_list.append(
            ap.models.AstroPhot_Model(
                name="basic sky",
                model_type="flat sky model",
                parameters={"F": -1},
                target=target,
            )
        )

        MODEL = ap.models.AstroPhot_Model(
            name="model",
            model_type="group model",
            target=target,
            models=model_list,
        )

        MODEL.initialize()

        res = ap.fit.Iter(MODEL, method=ap.fit.LM)

        res.fit()


class TestIterLM(unittest.TestCase):
    def test_iter_basic(self):
        target = make_basic_sersic()
        model_list = []
        model_list.append(
            ap.models.AstroPhot_Model(
                name="basic sersic",
                model_type="sersic galaxy model",
                parameters={
                    "center": [20, 20],
                    "PA": 60 * np.pi / 180,
                    "q": 0.5,
                    "n": 2,
                    "Re": 5,
                    "Ie": 1,
                },
                target=target,
            )
        )
        model_list.append(
            ap.models.AstroPhot_Model(
                name="basic sky",
                model_type="flat sky model",
                parameters={"F": -1},
                target=target,
            )
        )

        MODEL = ap.models.AstroPhot_Model(
            name="model",
            model_type="group model",
            target=target,
            models=model_list,
        )

        MODEL.initialize()

        res = ap.fit.Iter_LM(MODEL)

        res.fit()


class TestHMC(unittest.TestCase):
    def test_hmc_sample(self):
        np.random.seed(12345)
        N = 50
        pixelscale = 0.8
        true_params = {
            "n": 2,
            "Re": 10,
            "Ie": 1,
            "center": [-3.3, 5.3],
            "q": 0.7,
            "PA": np.pi / 4,
        }
        target = ap.image.Target_Image(
            data=np.zeros((N, N)),
            pixelscale=pixelscale,
        )

        MODEL = ap.models.Sersic_Galaxy(
            name="sersic model",
            target=target,
            parameters=true_params,
        )
        img = MODEL().data.detach().cpu().numpy()
        target.data = torch.Tensor(
            img
            + np.random.normal(scale=0.1, size=img.shape)
            + np.random.normal(scale=np.sqrt(img) / 10)
        )
        target.variance = torch.Tensor(0.1**2 + img / 100)

        HMC = ap.fit.HMC(MODEL, epsilon=1e-5, max_iter=5, warmup=2)
        HMC.fit()


class TestNUTS(unittest.TestCase):
    def test_nuts_sample(self):
        np.random.seed(12345)
        N = 50
        pixelscale = 0.8
        true_params = {
            "n": 2,
            "Re": 10,
            "Ie": 1,
            "center": [-3.3, 5.3],
            "q": 0.7,
            "PA": np.pi / 4,
        }
        target = ap.image.Target_Image(
            data=np.zeros((N, N)),
            pixelscale=pixelscale,
        )

        MODEL = ap.models.Sersic_Galaxy(
            name="sersic model",
            target=target,
            parameters=true_params,
        )
        img = MODEL().data.detach().cpu().numpy()
        target.data = torch.Tensor(
            img
            + np.random.normal(scale=0.1, size=img.shape)
            + np.random.normal(scale=np.sqrt(img) / 10)
        )
        target.variance = torch.Tensor(0.1**2 + img / 100)

        NUTS = ap.fit.NUTS(MODEL, max_iter=5, warmup=2)
        NUTS.fit()


class TestMHMCMC(unittest.TestCase):
    def test_singlesersic(self):
        np.random.seed(12345)
        N = 50
        pixelscale = 0.8
        true_params = {
            "n": 2,
            "Re": 10,
            "Ie": 1,
            "center": [-3.3, 5.3],
            "q": 0.7,
            "PA": np.pi / 4,
        }
        target = ap.image.Target_Image(
            data=np.zeros((N, N)),
            pixelscale=pixelscale,
        )

        MODEL = ap.models.Sersic_Galaxy(
            name="sersic model",
            target=target,
            parameters=true_params,
        )
        img = MODEL().data.detach().cpu().numpy()
        target.data = torch.Tensor(
            img
            + np.random.normal(scale=0.1, size=img.shape)
            + np.random.normal(scale=np.sqrt(img) / 10)
        )
        target.variance = torch.Tensor(0.1**2 + img / 100)

        MHMCMC = ap.fit.MHMCMC(MODEL, epsilon=1e-4, max_iter=100)
        MHMCMC.fit()

        self.assertGreater(
            MHMCMC.acceptance,
            0.1,
            "MHMCMC should have nonzero acceptance for simple fits",
        )


if __name__ == "__main__":
    unittest.main()



# No U-Turn Sampler variant of Hamiltonian Monte-Carlo
from typing import Optional, Sequence

import torch
import pyro
import pyro.distributions as dist
from pyro.infer import MCMC as pyro_MCMC
from pyro.infer import NUTS as pyro_NUTS
from pyro.infer.mcmc.adaptation import BlockMassMatrix
from pyro.ops.welford import WelfordCovariance

from .base import BaseOptimizer
from ..models import AstroPhot_Model

__all__ = ["NUTS"]


###########################################
# !Overwrite pyro configuration behavior!
# currently this is the only way to provide
# mass matrix manually
###########################################
def new_configure(self, mass_matrix_shape, adapt_mass_matrix=True, options={}):
    """
    Sets up an initial mass matrix.

    :param dict mass_matrix_shape: a dict that maps tuples of site names to the shape of
        the corresponding mass matrix. Each tuple of site names corresponds to a block.
    :param bool adapt_mass_matrix: a flag to decide whether an adaptation scheme will be used.
    :param dict options: tensor options to construct the initial mass matrix.
    """
    inverse_mass_matrix = {}
    for site_names, shape in mass_matrix_shape.items():
        self._mass_matrix_size[site_names] = shape[0]
        diagonal = len(shape) == 1
        inverse_mass_matrix[site_names] = (
            torch.full(shape, self._init_scale, **options)
            if diagonal
            else torch.eye(*shape, **options) * self._init_scale
        )
        if adapt_mass_matrix:
            adapt_scheme = WelfordCovariance(diagonal=diagonal)
            self._adapt_scheme[site_names] = adapt_scheme

    if len(self.inverse_mass_matrix.keys()) == 0:
        self.inverse_mass_matrix = inverse_mass_matrix


BlockMassMatrix.configure = new_configure
############################################


class NUTS(BaseOptimizer):
    """No U-Turn Sampler (NUTS) implementation for Hamiltonian Monte Carlo
    (HMC) based MCMC sampling.

    This is a wrapper for the Pyro package: https://docs.pyro.ai/en/stable/index.html

    The NUTS class provides an implementation of the No-U-Turn Sampler
    (NUTS) algorithm, which is a variation of the Hamiltonian Monte
    Carlo (HMC) method for Markov Chain Monte Carlo (MCMC)
    sampling. This implementation uses the Pyro library to perform the
    sampling. The NUTS algorithm utilizes gradients of the target
    distribution to more efficiently explore the probability
    distribution of the model.

    More information on HMC and NUTS can be found at:
    https://en.wikipedia.org/wiki/Hamiltonian_Monte_Carlo,
    https://arxiv.org/abs/1701.02434, and
    http://www.mcmchandbook.net/HandbookChapter5.pdf

    Args:
        model (AstroPhot_Model): The model which will be sampled.
        initial_state (Optional[Sequence], optional): A 1D array with the values for each parameter in the model. These values should be in the form of "as_representation" in the model. Defaults to None.
        max_iter (int, optional): The number of sampling steps to perform. Defaults to 1000.
        epsilon (float, optional): The step size for the NUTS sampler. Defaults to 1e-3.
        inv_mass (Optional[Tensor], optional): Inverse Mass matrix (covariance matrix) for the Hamiltonian system. Defaults to None.
        progress_bar (bool, optional): If True, display a progress bar during sampling. Defaults to True.
        prior (Optional[Distribution], optional): Prior distribution for the model parameters. Defaults to None.
        warmup (int, optional): Number of warmup (or burn-in) steps to perform before sampling. Defaults to 100.
        nuts_kwargs (Dict[str, Any], optional): A dictionary of additional keyword arguments to pass to the NUTS sampler. Defaults to {}.
        mcmc_kwargs (Dict[str, Any], optional): A dictionary of additional keyword arguments to pass to the MCMC function. Defaults to {}.

    Methods:
        fit(state: Optional[torch.Tensor] = None, nsamples: Optional[int] = None, restart_chain: bool = True) -> 'NUTS':
            Performs the MCMC sampling using a NUTS HMC and records the chain for later examination.

    """

    def __init__(
        self,
        model: AstroPhot_Model,
        initial_state: Optional[Sequence] = None,
        max_iter: int = 1000,
        **kwargs,
    ):
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        self.inv_mass = kwargs.get("inv_mass", None)
        self.epsilon = kwargs.get("epsilon", 1e-4)
        self.progress_bar = kwargs.get("progress_bar", True)
        self.prior = kwargs.get("prior", None)
        self.warmup = kwargs.get("warmup", 100)
        self.nuts_kwargs = kwargs.get("nuts_kwargs", {})
        self.mcmc_kwargs = kwargs.get("mcmc_kwargs", {})

    def fit(
        self,
        state: Optional[torch.Tensor] = None,
        nsamples: Optional[int] = None,
        restart_chain: bool = True,
    ):
        """
        Performs the MCMC sampling using a NUTS HMC and records the chain for later examination.
        """

        def step(model, prior):
            x = pyro.sample("x", prior)
            # Log-likelihood function
            model.parameters.flat_detach()
            log_likelihood_value = -model.negative_log_likelihood(
                parameters=x, as_representation=True
            )
            # Observe the log-likelihood
            pyro.factor("obs", log_likelihood_value)

        if self.prior is None:
            self.prior = dist.Normal(
                self.current_state,
                torch.ones_like(self.current_state) * 1e2 + torch.abs(self.current_state) * 1e2,
            )

        # Set up the NUTS sampler
        nuts_kwargs = {
            "jit_compile": False,
            "ignore_jit_warnings": True,
            "step_size": self.epsilon,
            "full_mass": True,
            "adapt_step_size": True,
            "adapt_mass_matrix": self.inv_mass is None,
        }
        nuts_kwargs.update(self.nuts_kwargs)
        nuts_kernel = pyro_NUTS(step, **nuts_kwargs)
        if self.inv_mass is not None:
            nuts_kernel.mass_matrix_adapter.inverse_mass_matrix = {("x",): self.inv_mass}

        # Provide an initial guess for the parameters
        init_params = {"x": self.model.parameters.vector_representation()}

        # Run MCMC with the NUTS sampler and the initial guess
        mcmc_kwargs = {
            "num_samples": self.max_iter,
            "warmup_steps": self.warmup,
            "initial_params": init_params,
            "disable_progbar": not self.progress_bar,
        }
        mcmc_kwargs.update(self.mcmc_kwargs)
        mcmc = pyro_MCMC(nuts_kernel, **mcmc_kwargs)

        mcmc.run(self.model, self.prior)
        self.iteration += self.max_iter

        # Extract posterior samples
        chain = mcmc.get_samples()["x"]

        with torch.no_grad():
            for i in range(len(chain)):
                chain[i] = self.model.parameters.vector_transform_rep_to_val(chain[i])
        self.chain = chain

        return self



import numpy as np


def fluxdens_to_fluxsum(R, I, axisratio):
    """
    Integrate a flux density profile

    R: semi-major axis length (arcsec)
    I: flux density (flux/arcsec^2)
    axisratio: b/a profile
    """

    S = np.zeros(len(R))
    S[0] = I[0] * np.pi * axisratio[0] * (R[0] ** 2)
    for i in range(1, len(R)):
        S[i] = trapz(2 * np.pi * I[: i + 1] * R[: i + 1] * axisratio[: i + 1], R[: i + 1]) + S[0]
    return S


def fluxdens_to_fluxsum_errorprop(
    R, I, IE, axisratio, axisratioE=None, N=100, symmetric_error=True
):
    """
    Integrate a flux density profile

    R: semi-major axis length (arcsec)
    I: flux density (flux/arcsec^2)
    axisratio: b/a profile
    """
    if axisratioE is None:
        axisratioE = np.zeros(len(R))

    # Create container for the monte-carlo iterations
    sum_results = np.zeros((N, len(R))) - 99.999
    I_CHOOSE = np.logical_and(np.isfinite(I), I > 0)
    if np.sum(I_CHOOSE) < 5:
        return (None, None) if symmetric_error else (None, None, None)
    sum_results[0][I_CHOOSE] = fluxdens_to_fluxsum(R[I_CHOOSE], I[I_CHOOSE], axisratio[I_CHOOSE])
    for i in range(1, N):
        # Randomly sampled SB profile
        tempI = np.random.normal(loc=I, scale=np.abs(IE))
        # Randomly sampled axis ratio profile
        tempq = np.clip(
            np.random.normal(loc=axisratio, scale=np.abs(axisratioE)),
            a_min=1e-3,
            a_max=1 - 1e-3,
        )
        # Compute COG with sampled data
        sum_results[i][I_CHOOSE] = fluxdens_to_fluxsum(
            R[I_CHOOSE], tempI[I_CHOOSE], tempq[I_CHOOSE]
        )

    # Condense monte-carlo evaluations into profile and uncertainty envelope
    sum_lower = sum_results[0] - np.quantile(sum_results, 0.317310507863 / 2, axis=0)
    sum_upper = np.quantile(sum_results, 1.0 - 0.317310507863 / 2, axis=0) - sum_results[0]

    # Return requested uncertainty format
    if symmetric_error:
        return sum_results[0], np.abs(sum_lower + sum_upper) / 2
    else:
        return sum_results[0], sum_lower, sum_upper


def _Fmode_integrand(t, parameters):
    fsum = sum(
        parameters["Am"][m] * np.cos(parameters["m"][m] * (t + parameters["Phim"][m]))
        for m in range(len(parameters["m"]))
    )
    dfsum = sum(
        parameters["m"][m]
        * parameters["Am"][m]
        * np.sin(parameters["m"][m] * (t + parameters["Phim"][m]))
        for m in range(len(parameters["m"]))
    )
    return (np.sin(t) ** 2) * np.exp(2 * fsum) + np.sin(t) * np.cos(t) * np.exp(fsum) * dfsum


def Fmode_Areas(R, parameters):
    A = []
    for i in range(len(R)):
        A.append((R[i] ** 2) * quad(_Fmode_integrand, 0, 2 * np.pi, args=(parameters[i],))[0])
    return np.array(A)


def Fmode_fluxdens_to_fluxsum(R, I, parameters, A=None):
    """
    Integrate a flux density profile, with isophotes including Fourier perturbations.

    Arguments
    ---------
    R: arcsec
      semi-major axis length

    I: flux/arcsec^2
      flux density

    parameters: list of dictionaries
      list of dictionary of isophote shape parameters for each radius.
      formatted as

      .. code-block:: python

        {
            "ellip": "ellipticity",
            "m": "list of modes used",
            "Am": "list of mode powers",
            "Phim": "list of mode phases",
        }

      entries for each radius.
    """
    if all(parameters[p]["m"] is None for p in range(len(parameters))):
        return fluxdens_to_fluxsum(
            R,
            I,
            1.0 - np.array(list(parameters[p]["ellip"] for p in range(len(parameters)))),
        )

    S = np.zeros(len(R))
    if A is None:
        A = Fmode_Areas(R, parameters)
    # update the Area calculation to be scaled by the ellipticity
    Aq = A * np.array(list((1 - parameters[i]["ellip"]) for i in range(len(R))))
    S[0] = I[0] * Aq[0]
    Adiff = np.array([Aq[0]] + list(Aq[1:] - Aq[:-1]))
    for i in range(1, len(R)):
        S[i] = trapz(I[: i + 1] * Adiff[: i + 1], R[: i + 1]) + S[0]
    return S


def Fmode_fluxdens_to_fluxsum_errorprop(R, I, IE, parameters, N=100, symmetric_error=True):
    """
    Integrate a flux density profile, with isophotes including Fourier perturbations.

    Arguments
    ---------
    R: arcsec
      semi-major axis length

    I: flux/arcsec^2
      flux density

    parameters: list of dictionaries
      list of dictionary of isophote shape parameters for each radius.
      formatted as

      .. code-block:: python

        {
            "ellip": "ellipticity",
            "m": "list of modes used",
            "Am": "list of mode powers",
            "Phim": "list of mode phases",
        }

      entries for each radius.
    """

    for i in range(len(R)):
        if "ellip err" not in parameters[i]:
            parameters[i]["ellip err"] = np.zeros(len(R))
    if all(parameters[p]["m"] is None for p in range(len(parameters))):
        return fluxdens_to_fluxsum_errorprop(
            R,
            I,
            IE,
            1.0 - np.array(list(parameters[p]["ellip"] for p in range(len(parameters)))),
            np.array(list(parameters[p]["ellip err"] for p in range(len(parameters)))),
            N=N,
            symmetric_error=symmetric_error,
        )

    # Create container for the monte-carlo iterations
    sum_results = np.zeros((N, len(R))) - 99.999
    I_CHOOSE = np.logical_and(np.isfinite(I), I > 0)
    if np.sum(I_CHOOSE) < 5:
        return (None, None) if symmetric_error else (None, None, None)
    cut_parameters = list(compress(parameters, I_CHOOSE))
    A = Fmode_Areas(R[I_CHOOSE], cut_parameters)
    sum_results[0][I_CHOOSE] = Fmode_fluxdens_to_fluxsum(
        R[I_CHOOSE], I[I_CHOOSE], cut_parameters, A
    )
    for i in range(1, N):
        # Randomly sampled SB profile
        tempI = np.random.normal(loc=I, scale=np.abs(IE))
        # Randomly sampled axis ratio profile
        temp_parameters = deepcopy(cut_parameters)
        for p in range(len(cut_parameters)):
            temp_parameters[p]["ellip"] = np.clip(
                np.random.normal(
                    loc=cut_parameters[p]["ellip"],
                    scale=np.abs(cut_parameters[p]["ellip err"]),
                ),
                a_min=1e-3,
                a_max=1 - 1e-3,
            )
        # Compute COG with sampled data
        sum_results[i][I_CHOOSE] = Fmode_fluxdens_to_fluxsum(
            R[I_CHOOSE], tempI[I_CHOOSE], temp_parameters, A
        )

    # Condense monte-carlo evaluations into profile and uncertainty envelope
    sum_lower = sum_results[0] - np.quantile(sum_results, 0.317310507863 / 2, axis=0)
    sum_upper = np.quantile(sum_results, 1.0 - 0.317310507863 / 2, axis=0) - sum_results[0]

    # Return requested uncertainty format
    if symmetric_error:
        return sum_results[0], np.abs(sum_lower + sum_upper) / 2
    else:
        return sum_results[0], sum_lower, sum_upper
